# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)


# django imports
from django.http import JsonResponse, HttpResponseBadRequest
from django.contrib.auth.decorators import permission_required
from django.db import models

# iop4lib imports
from iop4lib.db import AstroSource
from iop4lib.enums import SRCTYPES
from iop4lib.utils import qs_to_table

# other imports

#logging
import logging
logger = logging.getLogger(__name__)

# API VIEWS



@permission_required(["iop4api.view_astrosource"])
def catalog(request):
    qs = AstroSource.objects.exclude(srctype=SRCTYPES.CALIBRATOR)

    all_column_names = [f.name for f in AstroSource._meta.get_fields() if hasattr(f, 'verbose_name')]

    return JsonResponse(qs_to_table(qs=qs, column_names=all_column_names))



# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.db import models
from django.db.models import Count
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

# other imports
import math
import numpy as np
import astropy.units as u
from astropy.time import Time

# iop4lib imports
from iop4lib.enums import *
from .fields import FlagChoices, FlagBitField

# logging
import logging
logger = logging.getLogger(__name__)

class PhotoPolResultReducedFitRelation(models.Model):
    photopolresult = models.ForeignKey('PhotoPolResult', on_delete=models.CASCADE, related_name='reducedfits_relations')
    reducedfit = models.ForeignKey('ReducedFit', on_delete=models.CASCADE, related_name='photopolresults_relations')

    astrosource = models.ForeignKey('AstroSource', null=False, on_delete=models.CASCADE)
    reduction = models.CharField(max_length=20, choices=REDUCTIONMETHODS.choices, help_text="Reduction method used.")

    class Meta:
        app_label = 'iop4api'
        constraints = [
            models.UniqueConstraint(fields=['photopolresult', 'reducedfit', 'astrosource', 'reduction'], name='unique_photopolresultreducedfitrelation')
        ]

class PhotoPolResult(models.Model):
    """ Model for Photo-Polarimetric results, from a set of ReducedFits and for a given AstroSource, produced by a reduction method.

    .. note::
        A Photo Polarimetric will be uniquely identified by the reduced fits it had been built from, the source for which the result is, and the reduction method 
        used. Since the reducedfits are a ManyToManyField, we can not add a unique_together constraint to the combination of fields that is enforced in the DB level 
        (Django does not support this); instead we specify a custom m2m relationship and enforce the constraints there.
    
    .. note::
        Some fields like 'juliandate', 'band', 'exptime', etc are automatically derived from the reducedfits. This is done in two ways:
        - Every time an instance is saved, in PhotoPolResult.save(), through PhotoPolResult.clean().
        - Evert time a reducedfit is added or removed from the reducedfits ManyToManyField, through the m2m_changed signal handler in PhotoPolResult.

        These fields could perhaps be better implemented as properties, but having them as real fields simplifies things a lot in the admin or when filtering.

        The fields 'astrosource' and 'reduction' are derived from the m2m relationship PhotoPolResultReducedFitRelation also.

        None of these fields should be set manually, as they are automatically computed on save.
    """

    # DB fields

    ## identifiers

    reducedfits = models.ManyToManyField('ReducedFit', related_name='photopolresults', through='PhotoPolResultReducedFitRelation', through_fields=('photopolresult', 'reducedfit'))

    ## automaticaly computed when setting reducedfits

    astrosource = models.ForeignKey('AstroSource', null=True, on_delete=models.CASCADE, related_name='photopolresults', editable=False)
    reduction = models.CharField(null=True, max_length=20, choices=REDUCTIONMETHODS.choices, help_text="Reduction method used.", editable=False)
    juliandate = models.FloatField(null=True, help_text='Julian date of observation (mean of julian dates of all reducedfits). Automaticaly computed on save, do not edit.', editable=False)
    juliandate_min = models.FloatField(null=True, help_text='Minimum julian date of observation (min of julian dates of all reducedfits). Automaticaly computed on save, do not edit.', editable=False)
    juliandate_max = models.FloatField(null=True, help_text='Maximum julian date of observation (max of julian dates of all reducedfits). Automaticaly computed on save, do not edit.', editable=False)
    epoch = models.ForeignKey("Epoch", null=True, on_delete=models.CASCADE, help_text="Epoch of the observation.", editable=False)
    instrument = models.CharField(null=True, max_length=20, choices=INSTRUMENTS.choices, default=INSTRUMENTS.NONE, help_text="Instrument used for the observation.", editable=False)
    obsmode = models.CharField(null=True, max_length=20, choices=OBSMODES.choices, default=OBSMODES.NONE, help_text="Whether the observation was photometry or polarimetry.", editable=False)
    band = models.CharField(null=True, max_length=10, choices=BANDS.choices, default=BANDS.NONE, help_text="Band of the observation, as in the filter used (R, V, etc).", editable=False)
    exptime = models.FloatField(null=True, help_text="Exposure time in seconds.")

    ## photo-polarimetric reduction info

    aperpix = models.FloatField(null=True, help_text="Aperture radius in pixels.")
    bkg_flux_counts = models.FloatField(null=True, help_text="Background flux in counts from aperture photometry.")
    bkg_flux_counts_err = models.FloatField(null=True, help_text="Error for bkg_flux_counts.")
    flux_counts = models.FloatField(null=True, help_text="Flux in counts from aperture photometry (background-substracted).")
    flux_counts_err = models.FloatField(null=True, help_text="Error for flux_counts.")
    mag_inst = models.FloatField(null=True, help_text="Instrumental magnitude, computed directly from the flux counts.")
    mag_inst_err = models.FloatField(null=True, help_text="Error for mag_inst.")
    mag_zp = models.FloatField(null=True, help_text="Magnitude zero point (if it is a calibrator with known magnitude, it is the computed zero point from it, if it is not a calibrator, it is the zero point used to compute the magnitude of the source")
    mag_zp_err = models.FloatField(null=True, help_text="Error for mag_zp.")

    ## photo-polarimetric extra-info; e.g. to be used for automatic computation of instrumental polarization

    _x_px = models.FloatField(null=True, help_text="pixel position of the source in the image, x coordinate.")
    _y_px = models.FloatField(null=True, help_text="pixel position of the source in the image, y coordinate.")
    _q_nocorr = models.FloatField(null=True, help_text="value without correction for instrumental polarization!")
    _u_nocorr = models.FloatField(null=True, help_text="value without correction for instrumental polarization!")
    _p_nocorr = models.FloatField(null=True, help_text="value without correction for instrumental polarization!")
    _chi_nocorr = models.FloatField(null=True, help_text="value without correction for instrumental polarization!")

    ## photo-polarimetric results

    mag = models.FloatField(null=True, help_text="Magnitude of the source, result of the reduction.")
    mag_err = models.FloatField(null=True, help_text="Error for mag.")

    p = models.FloatField(null=True, help_text="Polarization of the source [0-1], result of the reduction.")
    p_err = models.FloatField(null=True, help_text="Error for p.")

    chi = models.FloatField(null=True, help_text="Polarization angle of the source [deg], result of the reduction.")
    chi_err = models.FloatField(null=True, help_text="Error for chi.")

    ## host galaxy correction 
    aperas = models.FloatField(null=True, help_text="Aperture radius in arcseconds.")
    mag_corr = models.FloatField(null=True, help_text="Magnitude corrected for host galaxy.")
    mag_corr_err = models.FloatField(null=True, help_text="Error for mag_corr.")
    p_corr = models.FloatField(null=True, help_text="Polarization corrected for host galaxy.")
    p_corr_err = models.FloatField(null=True, help_text="Error for p_corr.")
    used_mag_for_corr = models.FloatField(null=True, help_text="Magnitude used for host galaxy correction (when mag is not available).")
    used_mag_err_for_corr = models.FloatField(null=True, help_text="Error for used_mag_for_corr.")

    ## flags
    
    class FLAGS(FlagChoices):
        BAD_PHOTOMETRY = 1 << 1
        BAD_POLARIMETRY = 1 << 2

    flags = FlagBitField(choices=FLAGS.choices(), default=0, help_text="Flags for the quality of the result.")

    ## extra fields

    modified = models.DateTimeField(auto_now=True)

    class Meta:
        app_label = 'iop4api'
        verbose_name = 'Photo-Polarimetric Result'
        verbose_name_plural = 'Photo-Polarimetric Results'

    # repr and str

    def __repr__(self):
        return f'{self.__class__.__name__}.objects.get(id={self.id!r})'
    
    def __str__(self):
        return f'<{self.__class__.__name__} {self.id} | {self.obsmode} {self.band} {self.astrosource.name} JD {self.juliandate}>'
    
    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f'{self!r}')
        else:
            with p.group(4, f'<{self.__class__.__name__}(', ')>'):
                p.text(f'id: {self.id}')
                p.breakable()
                p.text(f"reducedfits: {list(self.reducedfits.values_list('id', flat=True))}")
                p.breakable()
                p.text(f'{self.obsmode} {self.band} {self.astrosource.name}')
                p.breakable()
                p.text(f'JD: {self.juliandate:.5f} ({Time(self.juliandate, format="jd").iso})')
                if self.mag is not None:
                    p.breakable()
                    p.text(f'mag: {self.mag:.3f} ± {self.mag_err:.3f}')
                if self.p is not None:
                    p.breakable()
                    p.text(f'p: {self.p:.3f} ± {self.p_err:.3f}')
                if self.chi is not None:
                    p.breakable()
                    p.text(f'chi: {self.chi:.3f} ± {self.chi_err:.3f}')    

    @classmethod
    def qs_exact_reducedfits(cls, reducedfits):
        """ Returns a queryset of PhotoPolResult with exactly the same set of reducedfits as the ones provided."""
        qs = cls.objects.annotate(reducedfits__count=Count('reducedfits')).filter(reducedfits__count=len(reducedfits))
        for reducedfit in reducedfits:
            qs = qs.filter(reducedfits=reducedfit)
        return qs


    @classmethod
    def create(cls, reducedfits, astrosource, reduction, **kwargs):
        """Creates a new instance of PhotoPolResult, for a set of ReducedFit objects, an AstroSource and a reduction method, or updates an existing one.
        
        Parameters
        ----------
        reducedfits : list of ReducedFit
            List of ReducedFit objects to be associated with the PhotoPolResult. They will be stored on the ManyToMany field 'reducedfits'.
        astrosource : AstroSource
            AstroSource object to be associated with the PhotoPolResult. It will be stored as additional information on the ManyToMany relationship with the 'reducedfits'.
        reduction : str
            Reduction method used. It will be stored as additional information on the ManyToMany relationship with the 'reducedfits'.

        Other Parameters
        ----------------
        Any other keyword argument will be set on the instance created, and saved.

        """

        photopolresult_qs = cls.qs_exact_reducedfits(reducedfits).filter(astrosource=astrosource, reduction=reduction)

        # if none found, we create one, else use existing

        if (photopolresult := photopolresult_qs.first()) is None:
            logger.debug(f'Creating DB entry photopolresult for {reducedfits=}, {astrosource=} and {reduction=}')
            photopolresult = cls(**kwargs)
            photopolresult.save() # we need to save it to use the manytomany field
            photopolresult.reducedfits.set(reducedfits, through_defaults={'astrosource': astrosource, 'reduction': reduction}, clear=True)
            photopolresult.save()
        else:
            logger.debug(f'Db entry for photopolresult already exists for {reducedfits=}, {astrosource=} and {reduction=}, using it instead.')

        # set kwargs fields

        for key, value in kwargs.items():
            setattr(photopolresult, key, value)
            photopolresult.save()

        # update automatically filled fields from reducedfits

        photopolresult.update_fields()

        # save and return

        photopolresult.save()

        return photopolresult

    # Methods that automatically compute fields from the reducedfits

    @classmethod 
    def get_juliandate_from_reducedfits(cls, reducedfits):
        return np.mean([reducedfit.juliandate for reducedfit in reducedfits])
    
    @classmethod
    def get_juliandate_min_from_reducedfits(cls, reducedfits):
        return np.min([reducedfit.juliandate for reducedfit in reducedfits])
    
    @classmethod
    def get_juliandate_max_from_reducedfits(cls, reducedfits):
        return np.max([reducedfit.juliandate for reducedfit in reducedfits])
    
    def get_astrosource(self):
        from .astrosource import AstroSource
        return AstroSource.objects.get(pk=self.reducedfits_relations.values_list('astrosource', flat=True).distinct().get())
    
    def get_reduction(self):
        return self.reducedfits_relations.values_list('reduction', flat=True).distinct().get()
    
    def get_juliandate(self):
        return PhotoPolResult.get_juliandate_from_reducedfits(self.reducedfits.all())
    
    def get_juliandate_min(self):
        return PhotoPolResult.get_juliandate_min_from_reducedfits(self.reducedfits.all())
    
    def get_juliandate_max(self):
        return PhotoPolResult.get_juliandate_max_from_reducedfits(self.reducedfits.all())
    
    def get_epoch(self):
        if not all([reducedfit.epoch == self.reducedfits.first().epoch for reducedfit in self.reducedfits.all()]):
            raise Exception('All reducedfits must have the same epoch')
        return self.reducedfits.first().epoch
    
    def get_band(self):
        if not all([reducedfit.band == self.reducedfits.first().band for reducedfit in self.reducedfits.all()]):
            raise Exception('All reducedfits must have the same band')
        return self.reducedfits.first().band
    
    def get_exptime(self):
        if not all([reducedfit.exptime == self.reducedfits.first().exptime for reducedfit in self.reducedfits.all()]):
            raise Exception('All reducedfits must have the same exptime')
        return self.reducedfits.first().exptime
    
    def get_obsmode(self):
        if not all([reducedfit.obsmode == self.reducedfits.first().obsmode for reducedfit in self.reducedfits.all()]):
            raise Exception('All reducedfits must have the same obsmode')
        return self.reducedfits.first().obsmode
    
    def get_instrument(self):
        if not all([reducedfit.instrument == self.reducedfits.first().instrument for reducedfit in self.reducedfits.all()]):
            raise Exception('All reducedfits must have the same instrument')
        return self.reducedfits.first().instrument
    
    def update_fields(self):
        self.astrosource = self.get_astrosource()
        self.reduction = self.get_reduction()
        self.juliandate = self.get_juliandate()
        self.juliandate_min = self.get_juliandate_min()
        self.juliandate_max = self.get_juliandate_max()
        self.epoch = self.get_epoch()
        self.band = self.get_band()
        self.exptime = self.get_exptime()
        self.obsmode = self.get_obsmode()
        self.instrument = self.get_instrument()

    def clean(self):
        if self.id: # .reducedfits can only be set after saving and therefore after having an id
            self.update_fields()
        else:
            pass
    
    def save(self, *args, **kwargs):
        """ Overriden to enforce clean() before saving. See PhotoPolResult.__docstring__ for more info."""
        self.clean()
        super().save(*args, **kwargs)
        
    # Host galaxy correction
        
    class NoHostCorrectionAvailable(Exception):
        pass

    def compute_host_galaxy_correction(self):
        r""" Computes the host galaxy correction and stores it in the appropriate fields in the DB.

        See [1] and the docstring of iop4lib.utils.get_host_correction for more info.

        References
        ----------
        [1] Nilsson, K., “Host galaxy subtraction of TeV candidate BL Lacertae objects”
            Astronomy and Astrophysics, vol. 475, no. 1, pp. 199-207, 2007. 
            doi:10.1051/0004-6361:20077624. 
            URL: https://ui.adsabs.harvard.edu/abs/2007A%26A...475..199N/abstract

        """

        from iop4lib.utils import get_host_correction

        if self.band != BANDS.R:
            raise PhotoPolResult.NoHostCorrectionAvailable('Host galaxy correction only available for R band')
        
        # compute the used aperture in arcseconds

        aperas = self.aperpix * self.reducedfits.first().pixscale.to(u.arcsec / u.pix).value
        
        # get the host galaxy flux for this aperture

        hostcorr_flux, hostcorr_flux_err = get_host_correction(self.astrosource, aperas)

        if hostcorr_flux is None:
            raise PhotoPolResult.NoHostCorrectionAvailable('No host galaxy correction available for this source')

        mag, mag_err, p, p_err = self.mag, self.mag_err, self.p, self.p_err

        # if there is no magnitude in this result, interpolate the lightcurve

        if mag is None:
            from iop4lib.utils import get_column_values
            qs = PhotoPolResult.objects.filter(astrosource=self.astrosource, band=self.band, mag__isnull=False, flags=0).order_by('juliandate')
            vals = get_column_values(qs, column_names=['juliandate', 'mag', 'mag_err'])
            mag, mag_err = np.interp(self.juliandate, vals['juliandate'], vals['mag']), np.interp(self.juliandate, vals['juliandate'], vals['mag_err'])
            used_mag_for_corr = mag
            used_mag_err_for_corr = mag_err
        else:
            used_mag_for_corr, used_mag_err_for_corr = None, None

        # compute the corrected magnitude and polarization
        # it is just a matter of substracting the host galaxy flux from the observed flux

        obsflux = 3.080e6  * 10 ** (-mag*0.4) # mag R to mJy
        obsflux_err = abs( 3.080e6 * 10**(-mag*0.4) * math.log(10) * 0.4 * mag_err )

        flux_corr = obsflux - hostcorr_flux
        flux_corr_err = math.sqrt( obsflux_err**2 + hostcorr_flux_err**2 )

        mag_corr = -2.5 * math.log10( flux_corr / (3.080e6) )
        mag_corr_err = abs( 2.5 * 1 / (flux_corr * math.log(10)) * flux_corr_err )

        if p is None:
            p_corr, p_corr_err = None, None
        else:
            p_corr = p * obsflux / flux_corr 
            p_corr_err = p_corr * math.sqrt( (p_err/p)**2 + (obsflux_err/obsflux)**2 + (flux_corr_err/flux_corr)**2 )

        # store results in DB and return

        self.mag_corr = mag_corr if used_mag_for_corr is None else None # if we used an interpolated magnitude, we don't store the corrected magnitude
        self.mag_corr_err = mag_corr_err if used_mag_for_corr is None else None

        self.p_corr = p_corr
        self.p_corr_err = p_corr_err

        self.aperas = aperas
        
        self.used_mag_for_corr = used_mag_for_corr
        self.used_mag_err_for_corr = used_mag_err_for_corr

        self.save()

        return mag_corr, p_corr, p_corr_err
        

@receiver(m2m_changed, sender=PhotoPolResult.reducedfits.through)
def reducedfits_on_change(sender, instance, action, **kwargs):
    """ Updates automatically filled fields when -after- reducedfits are altered."""
    if action.startswith('post_'):
        if instance.reducedfits.count() > 0:
            instance.update_fields()



"""
Django settings for iop4site project.

Generated by 'django-admin startproject' using Django 4.1.7.

For more information on this file, see
https://docs.djangoproject.com/en/4.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.1/ref/settings/
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


LOGIN_URL = "/iop4/login/"

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-t2%lj1okdbr#lvo37p**sw(#hoo$kd)n*jx=%r-%v33g&_!f2c"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True
INTERNAL_IPS = ["127.0.0.1"] ## when DEBUG=True, this enables the debug variable in templates

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    "iop4api",
    "iop4admin",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "iop4site.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "iop4site.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.1/ref/settings/#databases

import iop4lib.config
iop4conf = iop4lib.Config()

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": iop4conf.db_path,
    }
}


# Password validation
# https://docs.djangoproject.com/en/4.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

#USE_I18N = True
USE_I18N = False

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.1/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.1/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# FORMAT
USE_L10N = False
DATE_FORMAT = "Y-m-d"
DATE_INPUT_FORMATS = ["%Y-%m-%d"]

# LOGGING

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
    },
}


import pytest
from pathlib import Path

from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# django imports

# other imports
import os
from pytest import approx

# logging
import logging
logger = logging.getLogger(__name__)

@pytest.fixture
def load_test_catalog(testdata, django_db_setup, django_db_blocker):
    with django_db_blocker.unblock():
        from iop4lib.db import AstroSource

        # load test catalog in test db
        from django.core.management import call_command
        call_command('loaddata', str(Path(iop4conf.datadir) / 'testcatalog.yaml'), verbosity=0)



# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.http import JsonResponse, HttpResponseBadRequest
from django.contrib.auth.decorators import permission_required
from django.db import models

# iop4lib imports
from iop4lib.db import AstroSource, PhotoPolResult
from iop4lib.utils import qs_to_table

# other imports
from astropy.time import Time

#logging
import logging
logger = logging.getLogger(__name__)


@permission_required(["iop4api.view_photopolresult", "iop4api.view_astrosource"])
def data(request):

    source_name = request.POST.get("source_name", None)

    if not AstroSource.objects.filter(name=source_name).exists(): 
        return HttpResponseBadRequest(f"Source '{source_name}' does not exist".format(source_name=source_name))
    
    qs = PhotoPolResult.objects.filter(astrosource__name=source_name)

    data = qs.values()

    all_column_names = data[0].keys()
    default_column_names = ['juliandate', 'instrument', 'band', 'mag', 'mag_err', 'p', 'p_err', 'chi', 'chi_err']

    table_and_columns = qs_to_table(data=data, model=PhotoPolResult, column_names=all_column_names, default_column_names=default_column_names)
    
    result = {
                "data": table_and_columns["data"],
                "columns": table_and_columns["columns"],
                "query": {
                            "source_name": source_name, 
                            "count": qs.count()
                        }
    }

    # annotate with date fromt the julian date
    for r in result["data"]:
        r["date"] = Time(r["juliandate"], format='jd').iso
    result["columns"].append({"name": "date", "title": "date", "type": "date", "help": "date and time in ISO 8601 format, from the julian date"})    

    # annotate with flag labels
    for r in result["data"]:
        r["flag_labels"] = ",".join(PhotoPolResult.FLAGS.get_labels(r["flags"]))
    result["columns"].append({"name": "flag_labels", "title": "flag labels", "type": "string", "help": "flags as human readable labels"})

    return JsonResponse(result)




# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.http import StreamingHttpResponse
from django.contrib.auth.decorators import permission_required
from django.contrib.admin.views.decorators import staff_member_required

# iop4lib imports

# other imports
import coloredlogs
import coloredlogs.converter
from pathlib import Path

#logging
import logging
logger = logging.getLogger(__name__)


def get_log_file_generator(fpath):
    def _log_file_generator():
        with open(fpath, "r") as f:
            for line in f:
                yield coloredlogs.converter.convert(line)
    return _log_file_generator()

@staff_member_required
@permission_required(["iop4api.view_photpolresult", "iop4api.view_astrosource"]) 
def log(request):
    r"""Staff member required. Since the log file can be very large, we use a generator to stream it to the client."""

    if request.GET.get("log_file", None) is None:
        fpath = iop4conf.log_file
    else:
        fpath = str(Path(iop4conf.datadir) / "logs" / request.GET.get("log_file"))

    return StreamingHttpResponse(get_log_file_generator(fpath), content_type="text/plain")


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# other imports
import os
from pathlib import Path
import io
import base64
import numpy as np
import matplotlib as mplt
import matplotlib.pyplot as plt

# iop4lib
from iop4lib.db import AstroSource
from .singleobj import SingleObjView
from iop4lib.utils.plotting import plot_finding_chart

# logging
import logging
logger = logging.getLogger(__name__)

class AstroSourceDetailsView(SingleObjView):
    model = AstroSource
    template_name = "iop4admin/view_astrosourcedetails.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        
        obj = self.get_object()

        fields_and_values = {field.name:field.value_to_string(obj) for field in AstroSource._meta.fields if field.name != "comment" and getattr(obj, field.name) is not None}
        context['fields_and_values'] = fields_and_values

        # finding chart
        finding_char_path = Path(obj.filedpropdir) / "finding_chart.png"

        if not os.path.exists(finding_char_path) or iop4conf.iop4admin['force_rebuild_finding_charts']:
            buf = io.BytesIO()

            width, height = 800, 800

            fig = mplt.figure.Figure(figsize=(width/100, height/100), dpi=iop4conf.mplt_default_dpi)
            ax = fig.subplots()

            plot_finding_chart(obj, ax=ax, fig=fig)

            fig.savefig(buf, format='png', bbox_inches='tight')
            fig.clf()

            buf.seek(0)
            imgbytes = buf.read()

            if not os.path.exists(obj.filedpropdir):
                os.makedirs(obj.filedpropdir)
            with open(finding_char_path, 'wb') as f:
                f.write(imgbytes)
        else: 
            with open(finding_char_path, 'rb') as f:
                imgbytes = f.read()

        context['finding_chart_b64'] = base64.b64encode(imgbytes).decode('utf-8')

        return context
    



# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from abc import ABCMeta, abstractmethod

# other imports
import os
import re
import ftplib
from pathlib import Path
import astropy.io.fits as fits
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
import astrometry
import numpy as np
import math
import datetime

# iop4lib imports
from iop4lib.enums import *
from .telescope import Telescope, FTPArchiveMixin

# logging
import logging
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from iop4lib.db import RawFit, ReducedFit, Epoch


class OSNT090(FTPArchiveMixin, Telescope, metaclass=ABCMeta):
    
    # telescope identification

    name = "OSN-T090"
    abbrv = "T090"
    telescop_kw = "T90-OSN"

    # telescope specific properties

    ftp_address = iop4conf.osn_t090_address
    ftp_user = iop4conf.osn_t090_user
    ftp_password = iop4conf.osn_t090_password
    ftp_encoding = 'latin-1'

    re_expr_dirnames = re.compile(r"([0-9]{4}[0-9]{2}[0-9]{2})", flags=re.IGNORECASE)
    re_expr_fnames = re.compile('|'.join(iop4conf.osn_fnames_patterns), flags=re.IGNORECASE)

    # telescope specific methods

    @classmethod
    def check_telescop_kw(cls, rawfit):
        r""" Subclassed to account for DIPOL files, that may have empty TELESCOP keyword as of 2023-10-11 
        
        If it is empty, check first the instrument, and if it is DIPOL and the night is before 2023-10-11, then continue.

        Otherwise just call the parent method.
        """
        if rawfit.header["TELESCOP"] == "" and rawfit.night < datetime.date(2023, 10, 11):
            cls.classify_instrument_kw(rawfit)
            if rawfit.instrument == INSTRUMENTS.DIPOL:
                return
            
        super().check_telescop_kw(rawfit)

    # for the ftp archive mixin

    @classmethod 
    def remote_dirname_to_epochname(cls, dirname):
        matchs = cls.re_expr_dirnames.findall(dirname)
        if len(matchs) != 1:
            return None
        else:
            return f"{cls.name}/{matchs[0][0:4]}-{matchs[0][4:6]}-{matchs[0][6:8]}"

    @classmethod
    def epochname_to_remote_dirname(cls, epochname):
        from iop4lib.db import Epoch
        tel, night = Epoch.epochname_to_tel_night(epochname)
        return night.strftime("%Y%m%d")


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# iop4lib imports
from iop4lib.db import *
from iop4lib.enums import *
from iop4api.filters import *
from .singleobj import SingleObjView
from iop4admin.sites import iop4admin_site

# django imports
from django.urls import reverse

# other imports
import os
import base64
import glob
import numpy as np
from astropy.io import fits

# logging
import logging
logger = logging.getLogger(__name__)



def get_fit_view(viewcls, modelcls):
    class _(viewcls):
        model = modelcls
    return _


class FitJS9View(SingleObjView):

    template_name = "iop4admin/view_fitviewer.html"
    
    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)

        fit = self.get_object()

        context['fit'] = fit
        context['filesize_mb'] = f"{os.path.getsize(fit.filepath) / 1024 / 1024}"

        # urls

        context['fits_url'] = reverse(f'iop4admin:iop4api_{self.model._meta.model_name}_getfile', args=[fit.id])
        context['url_changelist'] = reverse('iop4admin:%s_%s_changelist' % (self.model._meta.app_label, self.model._meta.model_name)) + f"?id={fit.id}"

        if isinstance(fit, ReducedFit):
            context['url_raw'] = reverse('iop4admin:%s_%s_changelist' % (self.model._meta.app_label, fit.rawfit._meta.model_name)) + f"?id={fit.rawfit.id}"
        if isinstance(fit, RawFit) and hasattr(fit, "reduced"):
            context['url_reduced'] = reverse('iop4admin:%s_%s_changelist' % (self.model._meta.app_label, fit.reduced._meta.model_name)) + f"?id={fit.reduced.id}"

        return context
    


class FitDetailsView(SingleObjView):

    template_name = "iop4admin/view_fitdetails.html"

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)

        obj = self.get_object()

        # urls

        context['url_viewer'] = reverse(f"iop4admin:{self.model._meta.app_label}_{self.model._meta.model_name}_viewer", args=[obj.id])
        context['url_changelist'] = reverse('iop4admin:%s_%s_changelist' % (obj._meta.app_label, obj._meta.model_name)) + f"?id={obj.id}"

        if isinstance(obj, RawFit) and hasattr(obj, "reduced"):
            context['url_reduced'] = reverse('iop4admin:%s_%s_changelist' % (obj._meta.app_label, obj.reduced._meta.model_name)) + f"?id={obj.reduced.id}"

        if isinstance(obj, ReducedFit):
            context['url_raw'] = reverse('iop4admin:%s_%s_changelist' % (obj._meta.app_label, obj.rawfit._meta.model_name)) + f"?id={obj.rawfit.id}"

        ## non-empty fields and values
        fields_and_values = {field.name:str(getattr(obj, field.name)) for field in obj._meta.fields if getattr(obj, field.name) is not None}
        if hasattr(obj, 'flags'):
            fields_and_values['flags'] = ", ".join(obj.flag_labels)
        context['fields_and_values'] = fields_and_values

        # If file does not exist, return early

        if not obj.fileexists:
            return context
        
        # If file exists, gather additional info and pass it to the template
        
        imgdata = obj.mdata

        ## header

        with fits.open(obj.filepath) as hdul:
            context["header_L"] = [hdu.header for hdu in hdul]

        ## image preview parameters, image and histogram

        vmin = self.request.GET.get("vmin", "auto")
        vmax = self.request.GET.get("vmax", "auto")

        if vmin == "auto" or vmax == "auto":
            vmin = np.quantile(imgdata.compressed(), 0.3)
            vmax = np.quantile(imgdata.compressed(), 0.99)
            imgbytes = obj.get_imgbytes_preview_image()
        else:
            vmin = float(vmin)
            vmax = float(vmax)
            imgbytes = obj.get_imgbytes_preview_image(vmin=vmin, vmax=vmax)

        imgb64 = base64.b64encode(imgbytes).decode("utf-8")

        context["vmin"] = vmin
        context["vmax"] = vmax
        context['imgb64'] = imgb64

        histimg = obj.get_imgbytes_preview_histogram()
        histimg_b64 =  base64.b64encode(histimg).decode("utf-8")
        context['histimg_b64'] = histimg_b64
        
        # Reduction info

        ## astrometric calibration
        try:
            context["astrometry_info"] = obj.astrometry_info
            # logger.debug(f"Loaded astrometry info for {obj}: {context['astrometry_info']}")
        except Exception as e:
            # logger.debug(f"Failed to load astrometry info for {obj}: {e}")
            pass

        #logger.debug(f"Loading astrometry images from disk for {obj.fileloc}")

        astrometry_img_D = dict()
        fnameL = glob.iglob(obj.filedpropdir + "/astrometry_*.png")
        for fname in fnameL:
            with open(fname, "rb") as f:
                imgbytes = f.read()
            imgb64 = base64.b64encode(imgbytes).decode("utf-8")
            astrometry_img_D[os.path.basename(fname)] = imgb64
        context["astrometry_img_D"] = dict(sorted(astrometry_img_D.items()))

        #logger.debug(f"Finished loading images from disk for {obj.fileloc}")

        ## sources info
        if isinstance(obj, ReducedFit):
            sources_in_field_L = list()
            for source in obj.sources_in_field.all():
                sources_in_field_L.append((source, reverse('iop4admin:%s_%s_changelist' % (source._meta.app_label, source._meta.model_name)) + f"?id={source.id}"))
            context["sources_in_field_L"] = sources_in_field_L


        # stats
        context["stats"] = obj.stats

        #logger.debug("Finished building template context for fit details view")

        return context
    



# %% [markdown]
# # Step-by-step reduction

# The most common way to use IOP4 is by launching the `iop4` command to 
# automatically reduce some nights. However, you might be interested in reducing
# some images step by step (when debugging problems, or implementing new 
# instruments). In this notebook we will do exactly that, which is more or less 
# what `iop4.py` does when launched from the command line.

# %%
# %run 01_notebook_configuration.py

# %% [markdown]
# ## Fetching the data from the telescope archive.

# We might be interested in downloading all files from one night, or just a single 
# night. In any case, you can directly do

# %%
# warning: this might clean flags and some other fields:
from iop4lib.db import Epoch

epoch = Epoch.create(epochname="OSN-T090/2022-09-18")
epoch

# %% [markdown]

# When an epoch is created and `auto_link_rawfits=True`, IOP4 first attempts to
# list the remote telescope archive, then it reads the local archive. It tries
# to download missing files. If it cannot list the remote, but finds the
# local folder for that epoch, it will just give a warning. If it fails at
# both, it will return an error. This behavior can be tuned with 
# `check_remote_list`, `force_redownload` and `fallback_to_local` arguments to
# `create()`. Keep in mind that this will create the corresponding RawFit 
# instances if necessary.

# If you only want to fetch or create a raw fit instance you can equivalently do
# it with 

# %%
from iop4lib.db import RawFit
rf = RawFit.create(fileloc="OSN-T090/2022-09-18/BLLac-0001R.fit")
rf

# %% [markdown]
# Since this file was already created by the previous Epoch.create command, it
# gives a DEBUG message that the entry for this file already exists and re-uses 
# it.

# In this examples, `epochname` and `fileloc` are simply shorthands for  
# specifying `telescope`,`night` and `telescope`,`night`,`filename`,  
# in a way that matches the archive directory structure. These params uniquely 
# identify the observing epoch and the raw fits file, respectively.

# %% [markdown]
# ## Creating calibration frames

# Next step on a normal calibration process would be creating the master frames 
# for the current night. This can be done with 

# %% 
epoch.build_master_biases()
epoch.build_master_darks()
epoch.build_master_flats()

# %% [markdown]
# This will use all existing biases, darks and flats in `epoch` to create all 
# calibration frames (for each filter, etc) using as many images as possible.

# After this, we can create the reduced images for all images in the night

# %% 
epoch.reduce()

# %% [markdown]
# or just for our image of interest

# %% 
from iop4lib.db import ReducedFit
redf = ReducedFit.create(rawfit=rf)
redf

# %% [markdown]
# This will automatically select the appropriate master calibration frames from 
# its night (if they exist) or from the closest night. Otherwise, you can 
# specify which files to use by passing it as an argument to the creation 
# function.

# By default, ReducedFit.create will not create the reduced FITS file unless
# you explicitly pass `auto_build=True`. If you didn't, you can do it with

# %% 
redf.build_file()

# %% [markdown]
# This will create the reduced FITS file by applying the calibration frames to 
# the raw image, and will give it an appropriate WCS. To check the result, with 
# the astrometry included, you can use the utility function 
# `plot_preview_astrometry`:

# %%
from iop4lib.utils.plotting import plot_preview_astrometry
plot_preview_astrometry(redf)

# %% [markdown]
# In the images, we can see the coordiante frames the positions of the sources 
# in the catalog that appear in the image.

# Since the image it is a photometry image, we can directly compute 
# the relative photometry results with

# %%
redf.compute_relative_photometry()

from iop4lib.db import PhotoPolResult
PhotoPolResult.objects.filter(reducedfits__in=[redf]).first()

# %% [markdown]
# These results can already be inspected in the [web interface](/iop4/explore/plot/).


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports

# other imports
from abc import ABCMeta, abstractmethod

import re
import os
from pathlib import Path
import ftplib
import logging
import astropy.io.fits as fits
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
import astrometry
import numpy as np
import math
import datetime

# iop4lib imports
from iop4lib.enums import *

# logging
import logging
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from iop4lib.db import RawFit, ReducedFit, Epoch    

class Telescope(metaclass=ABCMeta):
    """ Base class for telescopes.

        Inherit this class to provide telescope specific functionality (e.g. discovering and 
        downloading new data, classification of instruments, etc).

        Attributes and methods that must be implemented are marked as abstract (they will give
        error if the class is inherited and the method is not implemented in the subclass).

        Other methods can be implemented in the subclass but are not required just raise NotImplementedError.

        Some classmethods are already defined since they are telescope independent; we should not need to
        override them (but it can be done).

        .. note::
            This class is abstract, it should not be instantiated.

        .. note::
            To add a new telescope, inherit this class and add it to the list of known telescopes
            in the get_known() method and the TelescopeEnum.
    """

    # Abstract attributes

    # these attributes must be implemented in the subclass

    # telescope identification

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def abbrv(self):
        pass

    @property
    @abstractmethod
    def telescop_kw(self):
        pass


    # Abstract methods

    # This methods must be implemented in the subclass

    @classmethod
    @abstractmethod
    def list_remote_raw_fnames(cls, epoch: 'Epoch') -> list[str] :
        pass

    @classmethod
    @abstractmethod
    def download_rawfits(cls, rawfits: list['RawFit']) -> None :
        pass

    @classmethod
    @abstractmethod
    def list_remote_epochnames(cls) -> list[str] :
        pass

    @classmethod
    @abstractmethod
    def list_remote_filelocs(cls, epochnames: list[str]) -> list[str] :
        pass

    # Class methods (you should be using these only from this Telescope class, not from subclasses)
    
    @classmethod
    def get_known(cls):
        from .cahat220 import CAHAT220
        from .osnt090 import OSNT090
        from .osnt150 import OSNT150
        return [CAHAT220, OSNT090, OSNT150]
    
    @classmethod
    def by_name(cls, name: str) -> 'Telescope':
        """
        Try to get telescope by name, then abbreviation, else raise Exception.
        """
        for tel in Telescope.get_known():
            if tel.name == name:
                return tel
        for tel in Telescope.get_known():
            if tel.abbrv == name:
                return tel
        raise NotImplementedError(f"Telescope {name} not implemented.")
        
    @classmethod
    def is_known(self, name):
        """
        Check if a telescope is known by name or abbreviation.
        """
        return (name in [tel.name for tel in Telescope.get_known()]) or (name in [tel.abbrv for tel in Telescope.get_known()])
    
    # telescope independent functionality
    # you should be using these from the subclasses already
    # these don't need to be overriden in subclasses, but they can be (e.g. OSN-T090 overrides check_telescop_kw)

    @classmethod
    def classify_rawfit(cls, rawfit: 'RawFit'):
        r""" Try to classify a RawFit object.

        This method will first check that the rawfit belongs to this telescope, 
        classify the instrument, then hand off classification to the instrument 
        class.        
        """

        from iop4lib.instruments import Instrument

        cls.check_telescop_kw(rawfit)
        cls.classify_instrument_kw(rawfit)
        Instrument.by_name(rawfit.instrument).classify_rawfit(rawfit)

    @classmethod
    def check_telescop_kw(cls, rawfit):
        """
        TELESCOP is an standard FITS keyword, it should not be telescope dependent.
        """
        telescop_header = fits.getheader(rawfit.filepath, ext=0)["TELESCOP"] 

        if (telescop_header != cls.telescop_kw):
            raise ValueError(f"TELESCOP in fits header ({telescop_header}) does not match telescope class kw ({cls.telescop_kw}).")
        
    @classmethod
    def classify_instrument_kw(cls, rawfit):
        """
        INSTRUME is an standard FITS keyword, it should not be telescope dependent.
        """

        instrume_header = fits.getheader(rawfit.filepath, ext=0)["INSTRUME"] 
        
        if instrume_header == "RoperT90" and rawfit.epoch.night < datetime.date(2021, 10, 23):
            # RoperT90 was replaced by AndorT90 on 2021-10-23, but the control PC was not updated until some time later
            rawfit.instrument = INSTRUMENTS.RoperT90
        elif instrume_header == "AndorT90" or (instrume_header == "RoperT90" and rawfit.epoch.night >= datetime.date(2021, 10, 23)):
            rawfit.instrument = INSTRUMENTS.AndorT90
        elif instrume_header == "Andor" or instrume_header == "AndorT150": # until 2023-01-11, AndorT150 was called simply Andor
            rawfit.instrument = INSTRUMENTS.AndorT150
        elif instrume_header == "CAFOS 2.2":
            rawfit.instrument = INSTRUMENTS.CAFOS
        elif instrume_header == "ASI Camera (1)":
            rawfit.instrument = INSTRUMENTS.DIPOL
        else:
            raise ValueError(f"INSTRUME in fits header ({instrume_header}) not known.")
    







class FTPArchiveMixin():

    @classmethod
    @abstractmethod
    def remote_dirname_to_epochname(cls, dirname):
        pass
        
    @classmethod
    @abstractmethod
    def epochname_to_remote_dirname(cls, epochname):
        pass


    @classmethod
    def list_remote_epochnames(cls):
        try:
            logger.debug(f"Loging to {cls.name} FTP server.")

            ftp =  ftplib.FTP(cls.ftp_address, encoding=cls.ftp_encoding)
            ftp.login(cls.ftp_user, cls.ftp_password)

            remote_dirnameL_all = ftp.nlst()
            ftp.quit()

            if '.' in remote_dirnameL_all:
                remote_dirnameL_all.remove('.')
            if '..' in remote_dirnameL_all:
                remote_dirnameL_all.remove('..')

            logger.debug(f"Total of {len(remote_dirnameL_all)} dirs in {cls.name} remote.")

            remote_epochnameL_all = list()

            for dirname in remote_dirnameL_all:
                if (epochname := cls.remote_dirname_to_epochname(dirname)) is None:
                    logger.warning(f"Could not parse {dirname} as a valid epochname.")
                else:
                    remote_epochnameL_all.append(epochname)

            logger.debug(f"Filtered to {len(remote_epochnameL_all)} epochs in {cls.name}.")

            return remote_epochnameL_all
        
        except Exception as e:
            raise Exception(f"Error listing remote epochs for {Telescope.name}: {e}.")

    @classmethod
    def list_remote_raw_fnames(cls, epoch):
        try:

            logger.debug(f"Loging to {cls.name} FTP server.")

            ftp =  ftplib.FTP(cls.ftp_address, encoding=cls.ftp_encoding)
            ftp.login(cls.ftp_user, cls.ftp_password)

            remote_dir = cls.epochname_to_remote_dirname(epoch.epochname)
            logger.debug(f"Trying to change {remote_dir} directory.")
            ftp.cwd(f"/{remote_dir}")

            remote_fnameL_all = ftp.nlst()
            ftp.quit()

            if '.' in remote_fnameL_all:
                remote_fnameL_all.remove('.')
            if '..' in remote_fnameL_all:
                remote_fnameL_all.remove('..')

            logger.debug(f"Total of {len(remote_fnameL_all)} files in {cls.name} {epoch.epochname}: {remote_fnameL_all}.")

            remote_fnameL = [s for s in remote_fnameL_all if cls.re_expr_fnames.search(s)]
            
            logger.debug(f"Filtered to {len(remote_fnameL)} files in {cls.name} {epoch.epochname}.")

            return remote_fnameL
        
        except Exception as e:
            raise Exception(f"Error listing remote dir for {epoch.epochname}: {e}.")
        
    @classmethod
    def download_rawfits(cls, rawfits: list['RawFit'] | list[str]):
        from iop4lib.db import RawFit

        try:
            ftp =  ftplib.FTP(cls.ftp_address, encoding=cls.ftp_encoding)
            ftp.login(cls.ftp_user, cls.ftp_password)

            for rawfit in rawfits:

                if isinstance(rawfit, str): # make sure that this is indeed the way tel, night, filepath are built
                    fileloc = rawfit
                    tel, night, filename = RawFit.fileloc_to_tel_night_filename(fileloc)
                    yyyy_mm_dd = night.strftime("%Y-%m-%d")
                    filepath = Path(iop4conf.datadir) / "raw" / tel / yyyy_mm_dd / filename
                elif isinstance(rawfit, RawFit):
                    fileloc = rawfit.fileloc
                    tel, night, filename = rawfit.telescope, rawfit.night, rawfit.filename
                    yyyy_mm_dd = night.strftime("%Y-%m-%d")
                    filepath =  Path(rawfit.filepath)

                rel_path = os.path.join(cls.epochname_to_remote_dirname(f"{tel}/{yyyy_mm_dd}"), filename)

                logger.debug(f"Downloading {fileloc} from {cls.name} archive ...")

                if not filepath.parent.exists():
                    os.makedirs(filepath.parent)

                with open(filepath, 'wb') as f:
                    ftp.retrbinary(f"RETR {rel_path}", f.write)

            ftp.quit()
        except Exception as e:
            raise Exception(f"Error downloading rawfits: {e}.")

    @classmethod
    def list_remote_filelocs(cls, epochnames: list[str]) -> list[str]:
        from iop4lib.db import Epoch

        ftp =  ftplib.FTP(cls.ftp_address, encoding=cls.ftp_encoding)
        ftp.login(cls.ftp_user, cls.ftp_password)

        dirnames = ftp.nlst()

        fileloc_list = list()

        for epochname in epochnames:

            tel, night = Epoch.epochname_to_tel_night(epochname)
            yymmdd = night.strftime("%y%m%d")

            remote_dir = cls.epochname_to_remote_dirname(epochname)

            if remote_dir not in dirnames:
                logger.error(f"{cls.name} remote dir {remote_dir} does not exist.")
                continue

            try:
                fnames = [fpath.split("/")[-1] for fpath in ftp.nlst(f"/{remote_dir}")]
                fileloc_list.extend([f"{epochname}/{fname}" for fname in fnames if cls.re_expr_fnames.search(fname) and fname != '.' and fname != '..'])
        

            except Exception as e:
                logger.error(f"Error listing {cls.name} remote dir for {epochname}: {e}.")
            
        ftp.quit()

        return fileloc_list


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# iop4lib imports
from iop4admin.sites import iop4admin_site

# django imports
from django.views.generic.detail import DetailView
from django.contrib.admin.views.decorators import staff_member_required
from django.utils.decorators import method_decorator

# other imports


# logging
import logging
logger = logging.getLogger(__name__)


@method_decorator(staff_member_required, name='dispatch')
class SingleObjView(DetailView):
    template_name = 'iop4admin/view_singleobj.html'  # specify your custom template here

    def get_object(self, queryset=None):
        obj = super().get_object(queryset)
        # Add any extra object loading or checking logic here
        return obj
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context.update(iop4admin_site.each_context(self.request))

        obj = self.get_object()

        context['title'] = f"{self.model.__name__} {obj.id}"
        context['app_label'] = obj._meta.app_label
        context['clsname'] = obj.__class__.__name__
        context['model_name'] = obj._meta.model_name
        context['model_verbose_name'] = obj._meta.verbose_name
        context['model_verbose_name_plural'] = obj._meta.verbose_name_plural

        return context


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# other imports
import os
import sys
import pytest
import yaml
import hashlib
from pathlib import Path

TEST_CONFIG = str(Path(iop4conf.datadir) / "config.tests.yaml")
TESTDATA_FPATH = str(Path("~/iop4testdata.tar.gz").expanduser())
TESTDATA_MD5SUM = '4d393377f8c659e2ead2fa252a9a38b2'
TEST_DATADIR = str(Path(iop4conf.datadir) / "iop4testdata")
TEST_DB_PATH = str(Path(iop4conf.db_path).expanduser().parent / ("test_" + str(Path(iop4conf.db_path).name)))

def pytest_configure():

    from django.conf import settings

    settings.configure(
        INSTALLED_APPS=[
            'iop4api',
        ],
        DATABASES = {
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": iop4conf.db_path,
                "TEST": {"NAME":TEST_DB_PATH},
            }
        },
        DEBUG = False,
    )
    
    iop4conf.configure(db_path=TEST_DB_PATH, datadir=TEST_DATADIR)
    iop4conf.config_path = TEST_CONFIG
    iop4conf.datadir = TEST_DATADIR
    iop4conf.db_path = TEST_DB_PATH
    
@pytest.fixture(scope="session")
def testdata(request):
    setUpClass()
    request.addfinalizer(tearDownClass)
    
def setUpClass():

    # check if test data is available
    if not os.path.exists(TESTDATA_FPATH):
        raise Exception("Test dataset not found")
    
    # check if test data is the right version
    with open(Path('~/iop4testdata.tar.gz').expanduser(), 'rb') as f:
        if TESTDATA_MD5SUM != hashlib.md5(f.read()).hexdigest():
            raise Exception("Test dataset version is not right (md5 sum mismatch)")

    # prepare test data dir
    try:
        os.makedirs(Path(iop4conf.datadir).parent, exist_ok=True)
    except Exception as e:
        raise Exception(f"Error creating test data directory: {e}")
    
    # unpack test data
    if os.system(f"tar -xzf {TESTDATA_FPATH} -C {Path(iop4conf.datadir).parent}") != 0:
        raise Exception("Error unpacking test dataset")
    
    # create test config file
    try:
        with open(TEST_CONFIG, 'w') as f:
            f.write(yaml.safe_dump(dict(iop4conf)))
    except Exception as e:
        raise Exception(f"Error creating test config file: {e}")

    # check if test data dir is the right one, else exist inmediately (to avoid deleting wrong dir)
    if Path(iop4conf.datadir).name != "iop4testdata":
        print("DANGER!")
        sys.exit(-1)

def tearDownClass():
    # check if test data dir is the right one, else exist inmediately (to avoid deleting wrong dir)
    if Path(iop4conf.datadir).name != "iop4testdata":
        print("DANGER!")
        sys.exit(-1)
    # remove test data dir
    os.system(f"rm -rf {iop4conf.datadir}")



import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)  

import numpy as np

from photutils.detection import DAOStarFinder
from astropy.convolution import convolve, convolve_fft
from photutils.segmentation import make_2dgaussian_kernel
from astropy.stats import SigmaClip, sigma_clipped_stats
from photutils.background import Background2D, MedianBackground, SExtractorBackground
from photutils.segmentation import detect_sources
from photutils.segmentation import SourceFinder
from photutils.segmentation import SourceCatalog


def get_bkg(imgdata, box_size=(16,16), filter_size=(11,11), mask=None):
    """Returns the 2D background for a given box_size and filter_size. Optionally, a mask can be provided (to mask sources or bad pixels)."""

    #bkg_estimator = MedianBackground()
    #bkg_estimator = MedianBackground(SigmaClip(sigma=3))

    bkg_estimator = SExtractorBackground() # default is to perform sigma clip with 3sigma and 5 iter
    #bkg_estimator = SExtractorBackground(SigmaClip(sigma=3.0))

    bkg = Background2D(imgdata, box_size, filter_size=filter_size, bkg_estimator=bkg_estimator, mask=mask)

    return bkg

def apply_gaussian_smooth(data, fwhm, kernel_size=None):
    if kernel_size is None:
        kernel_size = 2*int(fwhm)+1

    if kernel_size < 30:
        fconv = convolve
    else:
        fconv = convolve_fft

    kernel = make_2dgaussian_kernel(fwhm, size=kernel_size)
    data = fconv(data, kernel)
    return data

def get_sources_daofind(data, threshold=None, fwhm=8.0, n_threshold=5.0, brightest=100, exclude_border=True):
    """
    `data` needs not but should be bkg-substracted and smoothed (that is, convolved with a kernel),
    threshold should be set depending on the brackground noise.
    """

    mean, median, std = sigma_clipped_stats(data, sigma=5.0)

    if threshold is None:
        if n_threshold is not None:
            threshold = 5.0*std
        else:
            raise ValueError('threshold or n_threshold must be provided')

    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold, brightest=brightest, exclude_border=exclude_border)  
    sources = daofind(data-median)
    
    if sources is None or len(sources) == 0:
        return np.empty((0,0))
    
    sources.sort('flux', reverse=True)

    positions = np.transpose((sources['xcentroid'], sources['ycentroid']))
    
    return sources, positions

def get_cat_sources_from_segment_map(segment_map, imgdata_bkg_substracted, convolved_data):    
    """
    From a segmentation map, return a cataloge and the positions ((x1,y1), (x2,y2), ...).
    
    Needs also the bkg-substracted image and the convolved map to find the centroids from the segement map.
    """
    cat = SourceCatalog(data=imgdata_bkg_substracted, segment_img=segment_map, convolved_data=convolved_data)
    tb = cat.to_table()
    tb.sort('segment_flux', reverse=True)
    pos = np.transpose((tb['xcentroid'], tb['ycentroid']))
    return cat, pos, tb


def get_segmentation(imgdata_bkg_substracted, threshold, fwhm=1.0, kernel_size=None, npixels=64, deblend=True, mask=None):
    """
    Given the image with the background substracted, convolve it with a kernel (smooth it) and 
    find the segmentation map.
    
    Return the segemented map and the convolved image.
    """
    
    if kernel_size is None:
        kernel_size = 2*int(fwhm)+1

    if kernel_size < 30:
        fconv = convolve
    else:
        fconv = convolve_fft

    kernel = make_2dgaussian_kernel(fwhm, size=kernel_size)
    convolved_data = fconv(imgdata_bkg_substracted, kernel)

    if deblend:
        finder = SourceFinder(npixels=npixels, deblend=deblend, progress_bar=False)
        segment_map = finder(convolved_data, threshold)
    else:
        segment_map = detect_sources(convolved_data, threshold, npixels=npixels)

    if mask is not None:
        segment_map.remove_masked_labels(mask)

    return segment_map, convolved_data










# other functions

from sklearn.cluster import KMeans

def select_points(points, n, brightness=None):
    """ returns n points approximately uniformly distributed over the image, selecting the closes one to each cluster or the brightest one if brightness is provided."""

    # Convert the list of tuples to a 2D numpy array
    points_arr = np.array(points)

    # Create a KMeans instance
    kmeans = KMeans(n_clusters=n, random_state=0, n_init=10)

    # Perform the clustering
    kmeans.fit(points_arr)

    # Get labels for all points
    labels = kmeans.labels_

    selected_points = []
    selected_points_idx = []
    clusters = []
    for i in range(n):
        # Get the points in this cluster
        cluster_points = points_arr[labels == i]

        if brightness is not None:
            # Get the brightness values for these points
            cluster_brightness = np.array(brightness)[labels == i]

            # Find the index of the brightest point
            brightest_idx = np.argmax(cluster_brightness)
        else:
            brightest_idx = np.argmin(np.linalg.norm(points_arr[labels == i]-kmeans.cluster_centers_[:, np.newaxis]))

        # Add the brightest point to the list
        selected_points.append(cluster_points[brightest_idx])
        ## add also the index 
        selected_points_idx.append(np.argwhere(labels == i)[brightest_idx][0])

        # Append the cluster points to the clusters list
        clusters.append(cluster_points.tolist())

    # Return the brightest points and the clusters
    return selected_points, selected_points_idx, clusters, kmeans


# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.shortcuts import render
from django.urls import reverse
from django.shortcuts import render, redirect

# iop4lib imports
from iop4lib.db import AstroSource
from iop4lib.enums import SRCTYPES

# other imports
import json
import os
import subprocess
from pathlib import Path

#logging
import logging
logger = logging.getLogger(__name__)

GIT_COMMIT_HASH = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=os.path.dirname(os.path.realpath(__file__))).decode('ascii').strip()
GIT_BRANCH = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'], cwd=os.path.dirname(os.path.realpath(__file__))).decode('ascii').strip()
GIT_DESCRIBE = subprocess.check_output(['git', 'describe', '--always'], cwd=os.path.dirname(os.path.realpath(__file__))).decode('ascii').strip()

def index(request, tabs=None):

    context = {}

    tab_tree = {"about":{}, "login":{}, "explore": {"catalog":{}, "query":{}, "plot":{}, "data":{}, "logs":{}}}
    context["tab_tree"] =  json.dumps(tab_tree)
    
    # pass the tabs to the template (e.g. /iop4/tab1/tab2/, see urls.py)
    if tabs is not None:
        context['tabs'] = {f"C{i+1}selectedTab":tab for i, tab in enumerate(tabs)}

        # redirect to login if they are trying to see a tab that requires login
        if not request.user.is_authenticated and "C1selectedTab" in context['tabs'] and context['tabs']["C1selectedTab"] not in ["about", "login"]:
            return redirect("{}".format(reverse('iop4api:index', args=[["login",]])))
        
        # # # if the tab is not in the tab tree, redirect to the index
        # for i, tab in enumerate(tabs):
        #     if i == 0 and tab not in tab_tree.keys():
        #         return redirect("{}".format(reverse('iop4api:index')))

    # if the user is logged, pass source names to the template
    if request.user.is_authenticated:
        context['source_name_list'] = AstroSource.objects.exclude(srctype=SRCTYPES.CALIBRATOR).values_list('name', flat="True")

    # add the hash of the current commit installed 
    context['git_commit_hash'] = GIT_COMMIT_HASH
    context['git_branch'] = GIT_BRANCH
    context['git_describe'] = GIT_DESCRIBE

    # add the available log files
    context['log_files'] = json.dumps(sorted([os.path.basename(f) for f in os.listdir(Path(iop4conf.datadir) / "logs") if f.endswith(".log")]))

    return render(request, 'iop4api/index.html', context)



from django.contrib import admin

from django.utils.html import format_html
from django.urls import reverse 
from django.utils.safestring import mark_safe
from iop4api.filters import *
from iop4api.models import *
from .fitfile import AdminFitFile, action_mark_ignore, action_unmark_ignore

import logging
logger = logging.getLogger(__name__)

class AdminMasterBias(AdminFitFile):

    model = MasterBias

    list_display = ['id', 'telescope', 'night', 'instrument', 'imgsize', 'get_built_from', 'options', 'status']

    list_filter = (
            RawFitIdFilter,
            RawFitTelescopeFilter,
            RawFitNightFilter,
            RawFitInstrumentFilter,
            RawFitFlagFilter,
            "imgsize",
    )
    
    actions = [action_mark_ignore, action_unmark_ignore]
    
    @admin.display(description='Options')
    def options(self, obj):
        url_details = reverse(f"iop4admin:{self.model._meta.app_label}_{self.model._meta.model_name}_details", args=[obj.id])
        url_viewer = reverse(f"iop4admin:{self.model._meta.app_label}_{self.model._meta.model_name}_viewer", args=[obj.id])
        return format_html(rf'<a href="{url_details}">details</a> / <a href="{url_viewer}">advanced viewer</a>')
    
    @admin.display(description='Telescope')
    def telescope(self, obj):
        return obj.epoch.telescope
    
    @admin.display(description='Night')
    def night(self, obj):
        return obj.epoch.night
    


#!/usr/bin/env python
""" iop4.py
    Invoke the IOP4 pipeline.

You can specify how to select epochs using the --epoch-list, --discover-missing and --list-local options.
You can check whether to keep selected epochs that are already in DB using the --no-check-db option, otherwise
they will be removed from the list of epochs to process. The --date-start and --date-end options allow to filter
the selected epochs by date. 

Equivalent options exist for files.

Use -o option=value to override config options, e.g., to set the log level to DEBUG and use 6 parallel processes, 
you would invoke iop4 as `iop4 -o log_level=DEBUG -o n_processes=6 [other options]`.

Contact: Juan Escudero Pedrosa (jescudero@iaa.es).
"""

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=True)

#other imports
import os
import sys
import argparse
import coloredlogs
import numpy as np
import scipy as sp
import pandas as pd
import matplotlib as mplt
import matplotlib.pyplot as plt
import itertools
from astropy.time import Time
import datetime
import time
from pathlib import Path

# iop4lib imports
from iop4lib.db import *
from iop4lib.enums import *
from iop4lib.telescopes import *
from iop4lib.instruments import *

# logging
import logging
logger = logging.getLogger(__name__)



def process_epochs(epochname_list, force_rebuild, check_remote_list):

    epoch_L  : list[Epoch] = list()
    
    logger.info("Epochs will be created.")

    for epochname in epochname_list:
        epoch = Epoch.create(epochname=epochname, check_remote_list=check_remote_list)
        epoch_L.append(epoch)

    logger.info("Creating Master Biases")

    for epoch in epoch_L:
        
        epoch.build_master_biases(force_rebuild=force_rebuild)

    logger.info("Creating Master Darks.")

    for epoch in epoch_L:
        epoch.build_master_darks(force_rebuild=force_rebuild)
        
    logger.info("Creating Master Flats.")

    for epoch in epoch_L:
        epoch.build_master_flats(force_rebuild=force_rebuild)

    logger.info("Science files will be reduced.")

    rawfits = RawFit.objects.filter(epoch__in=epoch_L, imgtype=IMGTYPES.LIGHT).all()
    Epoch.reduce_rawfits(rawfits.filter(obsmode=OBSMODES.PHOTOMETRY), force_rebuild=force_rebuild)
    Epoch.reduce_rawfits(rawfits.filter(obsmode=OBSMODES.POLARIMETRY), force_rebuild=force_rebuild)

    logger.info("Computing results.")

    for epoch in epoch_L:
        PhotoPolResult.objects.filter(epoch=epoch).delete()
        epoch.compute_relative_photometry()
        epoch.compute_relative_polarimetry()

    logger.info("Applying corrections.")

    for epoch in epoch_L:
        results = PhotoPolResult.objects.filter(epoch=epoch).all()
        for result in results:
            try:
                result.compute_host_galaxy_correction()
            except PhotoPolResult.NoHostCorrectionAvailable:
                pass
            except Exception as e:
                logger.error(f"Error computing host galaxy correction for {result}: {e}")

    logger.info("Done.")


def list_local_epochnames():
    """ List all local epochnames in local archives (by looking at the raw directory)."""

    local_epochnames = list()

    for tel_cls in Telescope.get_known():
        if os.path.isdir(f"{iop4conf.datadir}/raw/{tel_cls.name}/"):
            local_epochnames.extend([f"{tel_cls.name}/{night}" for night in os.listdir(f"{iop4conf.datadir}/raw/{tel_cls.name}/")])

    return local_epochnames
    
def list_remote_epochnames():
    """ List all remote epochnames in remote archives.
    """
    epochnames = list()

    for tel_cls in Telescope.get_known():
        epochnames.extend(tel_cls.list_remote_epochnames())

    return epochnames


def discover_missing_epochs():
    """ Discover missing epochs in remote archive."""
    return list(set(list_remote_epochnames()).difference(list_local_epochnames()))    


def list_remote_filelocs(epochnames: None | list[str] = None):
    """ Discover files in remote archive for the given epochs.
    
    Use this function to list all files in the remote archive for the given epochs.
    It avoids calling list_remote_raw_fnames() for each epoch.
    """
    
    if epochnames is None:
         epochnames = list_remote_epochnames()

    grouped_epochnames = group_epochnames_by_telescope(epochnames)

    filelocs = list()

    for tel, epochnames in grouped_epochnames.items():
        if len(epochnames) > 0:
            filelocs.extend(Telescope.by_name(tel).list_remote_filelocs(epochnames))

    return filelocs

def list_local_filelocs():
    """ Discover local filelocs in local archive."""

    local_filelocs = list()

    for tel_cls in Telescope.get_known():
        if os.path.isdir(f"{iop4conf.datadir}/raw/{tel_cls.name}/"):
            for d in os.scandir(f"{iop4conf.datadir}/raw/{tel_cls.name}/"):
                local_filelocs.extend([f"{tel_cls.name}/{d.name}/{f.name}" for f in os.scandir(f"{iop4conf.datadir}/raw/{tel_cls.name}/{d.name}")])

    return local_filelocs

def discover_missing_filelocs():
    """ Discover missing files in remote archive.

    Compares the lists of remote files with the list of local files and returns the fileloc of the missing files.

    If epochnames is None, all remote epochs will be checked.
    """

    return list(set(list_remote_filelocs()).difference(list_local_filelocs()))


def retry_failed_files():
    qs = ReducedFit.objects.filter(flags__has=ReducedFit.FLAGS.ERROR_ASTROMETRY).all()
    logger.info(f"Retrying {qs.count()} failed reduced fits.")
    Epoch.reduce_reducedfits(qs)
    qs2 = ReducedFit.objects.filter(flags__has=ReducedFit.FLAGS.ERROR_ASTROMETRY).all()
    logger.info(f"Fixed {qs.count()-qs2.count()} out of {qs.count()} failed reduced fits.")



def filter_epochname_by_date(epochname_list, date_start=None, date_end=None):
    if date_start is not None:
        epochname_list = [epochname for epochname in epochname_list if Epoch.epochname_to_tel_night(epochname)[1] >= datetime.date.fromisoformat(date_start)]
    
    if date_end is not None:
        epochname_list = [epochname for epochname in epochname_list if Epoch.epochname_to_tel_night(epochname)[1] <= datetime.date.fromisoformat(date_end)]
    
    return epochname_list

def filter_filelocs_by_date(fileloc_list, date_start=None, date_end=None):

    if date_start is not None:
        fileloc_list = [fileloc for fileloc in fileloc_list if RawFit.fileloc_to_tel_night_filename(fileloc)[1] >= datetime.date.fromisoformat(date_start)]
    
    if date_end is not None:
        fileloc_list = [fileloc for fileloc in fileloc_list if RawFit.fileloc_to_tel_night_filename(fileloc)[1] <= datetime.date.fromisoformat(date_end)]
    
    return fileloc_list


def group_epochnames_by_telescope(epochnames):
    
        epochnames_by_telescope = {tel_cls.name:list() for tel_cls in Telescope.get_known()}
    
        for epochname in epochnames:
            tel, night = Epoch.epochname_to_tel_night(epochname)
            epochnames_by_telescope[tel].append(epochname)
    
        return epochnames_by_telescope

def group_filelocs_by_telescope(filelocs):

    filelocs_by_telescope = {tel_cls.name:list() for tel_cls in Telescope.get_known()}

    for fileloc in filelocs:
        tel, night, filename = RawFit.fileloc_to_tel_night_filename(fileloc)
        filelocs_by_telescope[tel].append(fileloc)

    return filelocs_by_telescope


def parse_config_overrides(overrides):

    config = dict()

    for override in overrides:
        key, value = override.split('=')
        config[key] = value

    # correct the type of the config options
    for k, v in config.items():
        # interpret the log_level option
        if k == "log_level" and not v.isnumeric():
            v = config[k] = getattr(logging, v.upper())
        # check the type of the config option, if it is an int, convert it
        if k in iop4conf and isinstance(iop4conf[k], int):
            config[k] = int(v)
            
    return config

def main():
    # Parse args:

    parser = argparse.ArgumentParser(
                    prog="iop4",
                    description="The Interactive Optical Photo-Polarimetric Python Pipeline",
                    epilog = __doc__,
                    formatter_class=argparse.RawDescriptionHelpFormatter,
                    allow_abbrev=False)
    
    parser.add_argument('-i', "--interactive", dest="interactive", action="store_true", help="<Optional> Jump to an IPython shell after finishing execution")
    parser.add_argument('--check-config', action="store_true", help="<Optional> Check the current configuration and exit")
    parser.add_argument('-o', action='append', dest='config_overrides', help="Override a config option (e.g., -o option=value)")
    
    # epoch processing options
    parser.add_argument('--epoch-list', dest='epochname_list', nargs='+', help='<Optional> List of epochs (e.g: T090/230102 T090/230204)')
    parser.add_argument('--discover-missing', action='store_true', help='<Optional> Discover new epochs to process them')
    parser.add_argument('--list-local', action='store_true', help='<Optional> Discover local epochs to process them')
    parser.add_argument('--list-only', action='store_true', help='<Optional> If given, the built list of epochs will be printed but not processed')
    parser.add_argument('--no-check-db', dest='keep_epochs_in_db', action='store_true', help='<Optional> Process discovered epochs even if they existed in archive')

    ## file processing  options
    parser.add_argument('--file-list', dest='fileloc_list', nargs='+', help='<Optional> List of files (e.g: tel/yyyy-mm-dd/name))')
    parser.add_argument('--discover-missing-files', action='store_true', help='<Optional> Discover files in remote archives that are not present in archive')
    parser.add_argument('--list-local-files', action='store_true', help='<Optional> Discover local files to process them')
    parser.add_argument('--list-files-only', action='store_true', help='<Optional> If given, the built list of filelocs will be printed but not processed')
    parser.add_argument('--no-check-db-files',  dest='keep_files_in_db', action='store_true', help='<Optional> Process discovered files even if they existed in archive')
    

    # other options
    parser.add_argument('--skip-remote-file-list', action='store_true', help='<Optional> Skip remote file list check')
    parser.add_argument("--force-rebuild", action="store_true", help="<Optional> Force re-building of files (pass force_rebuild=True)")
    parser.add_argument('--retry-failed', action='store_true', help='<Optional> Retry failed reduced fits')
    parser.add_argument('--reclassify-rawfits', dest="reclassify_rawfits", action="store_true", help="<Optional> Re-classify rawfits")

    # range
    parser.add_argument('--date-start', '-s', type=str, default=None, help='<Optional> Start date (YYYY-MM-DD)')
    parser.add_argument('--date-end', '-e', type=str, default=None, help='<Optional> End date (YYYY-MM-DD)')

    args = parser.parse_args()

    if len(sys.argv) < 2:
        parser.print_usage()
        sys.exit(0)

    # Configuration:

    if args.config_overrides:
        # get the config options as a dict
        opts = parse_config_overrides(args.config_overrides)
        # update the config
        iop4conf.update(opts)

    if not iop4conf.is_valid():
        print("check_config: there are some missing keys in the config file. All keys in the example config file must be present. Add them to your config file (you can set them to null) and try again. Aborting.")
        sys.exit(-1)
    elif args.check_config:
        print("check_config: config file is OK.")
        sys.exit(0)

    ## configure logging

    ROOT_LOGGER = logging.getLogger()
    
    ROOT_LOGGER.setLevel(iop4conf.log_level)

    logger_h1 = logging.FileHandler(iop4conf.log_file, mode="w")
    logger_h1.setFormatter(coloredlogs.ColoredFormatter(iop4conf.log_format, datefmt=iop4conf.log_date_format))

    logger_h2 = logging.StreamHandler(sys.stdout)
    logger_h2.setFormatter(coloredlogs.ColoredFormatter(iop4conf.log_format, datefmt=iop4conf.log_date_format))

    ROOT_LOGGER.handlers.clear()
    ROOT_LOGGER.addHandler(logger_h1)
    ROOT_LOGGER.addHandler(logger_h2)

    # Epochs
    
    epochnames_to_process = set()

    if args.list_local:
        logger.info("Listing local epochs...")
        local_epochs = list_local_epochnames()
        epochnames_to_process = epochnames_to_process.union(local_epochs)
        logger.info(f"Listed {len(local_epochs)} local epochs.")

    if args.discover_missing:
        logger.info("Discovering missing epochs...")
        missing_epochs = discover_missing_epochs()
        epochnames_to_process = epochnames_to_process.union(missing_epochs)
        logger.info(f"Discovered {len(missing_epochs)} missing epochs.")

    if args.epochname_list is not None:
         logger.info("Adding epochs from command line...")
         epochnames_to_process = epochnames_to_process.union(args.epochname_list)
         logger.info(f"Added {len(args.epochname_list)} epochs from command line.")

    if len(epochnames_to_process) > 0 and not args.keep_epochs_in_db:
        logger.info("Removing epochs already in the DB...")
        epochnames_in_db = set([epoch.epochname for epoch in Epoch.objects.all()])
        epochnames_to_process = epochnames_to_process.difference(epochnames_in_db)
        logger.info(f"Left {len(epochnames_to_process)} epochs to process.")

    logger.info(f"Gathered {len(epochnames_to_process)} epochs to process between {args.date_start} and {args.date_end}.")

    if args.date_start is not None or args.date_end is not None:
        logger.info("Filtering epochs by date.")
        epochnames_to_process = filter_epochname_by_date(epochnames_to_process, args.date_start, args.date_end)
        logger.info(f"Filtered to {len(epochnames_to_process)} epochs to process between {args.date_start} and {args.date_end}.")

    logger.debug(f"{epochnames_to_process=}")

    if not args.list_only:
        if len(epochnames_to_process) > 0:
            logger.info("Processing epochs.")
            epochnames_to_process, _ = list(zip(*sorted([(epochname, Epoch.epochname_to_tel_night(epochname)[1].strftime("%Y-%m-%d")) for epochname in epochnames_to_process], key=lambda x: x[1], reverse=True)))
            process_epochs(epochnames_to_process, args.force_rebuild, check_remote_list=~args.skip_remote_file_list)
    else:
        logger.info("Invoked with --list-only!")

    # Files

    filelocs_to_process = set()

    if args.list_local_files:
        logger.info("Listing local files...")
        filelocs_local = list_local_filelocs()
        filelocs_to_process = filelocs_to_process.union(filelocs_local)
        logger.info(f"Listed {len(filelocs_local)} local files.")
    else:
        filelocs_local = set()

    if args.discover_missing_files:
        logger.info("Discovering missing files...")
        filelocs_missing = discover_missing_filelocs()
        filelocs_to_process = filelocs_to_process.union(filelocs_missing)
        logger.info(f"Discovered {len(filelocs_missing)} missing files.")
    else:
        filelocs_missing = set()

    if args.fileloc_list is not None:
        logger.info("Adding files from command line...")
        filelocs_to_process = filelocs_to_process.union(args.file_list)
        logger.info(f"Added {len(args.file_list)} files from command line.")

    if len(filelocs_to_process) > 0 and not args.keep_files_in_db:
        logger.info(f"Removing files already in the DB ({RawFit.objects.count()}).")
        filelocs_in_db = set([rawfit.fileloc for rawfit in RawFit.objects.all()])
        filelocs_to_process = filelocs_to_process.difference(filelocs_in_db)
        logger.info(f"Left {len(filelocs_to_process)} files to process.")

    logger.info(f"Gathered {len(filelocs_to_process)} files to process.")

    if args.date_start is not None or args.date_end is not None:
        logger.info("Filtering files by date...")
        filelocs_to_process = filter_filelocs_by_date(filelocs_to_process, args.date_start, args.date_end)    
        logger.info(f"Filtered to {len(filelocs_to_process)} filelocs_to_process  between {args.date_start} and {args.date_end}.")

        filelocs_missing = set(filelocs_to_process).intersection(filelocs_missing)

    logger.debug(f"{filelocs_to_process=}")

    if not args.list_files_only:

        if args.discover_missing_files and len(filelocs_missing) > 0:
            logger.info("Downloading missing files.")
            for telname, filelocs in group_filelocs_by_telescope(filelocs_missing).items():
                Telescope.by_name(telname).download_rawfits(filelocs)

        if len(filelocs_to_process) > 0:
            logger.info("Processing files.")
            for fileloc in filelocs_to_process:
                rawfit = RawFit.create(fileloc=fileloc)

    else:
        logger.info("Invoked with --list-files-only")

    # Classify rawfits if indicated

    if args.reclassify_rawfits:
        
        logger.info("Classifying rawfits.")

        for epochname in epochnames_to_process:
            epoch = Epoch.by_epochname(epochname)
            for rawfit in epoch.rawfits.all():
                rawfit.classify()

        for fileloc in filelocs_to_process:
            rawfit = RawFit.by_fileloc(fileloc)
            rawfit.classify()

    # Retry failed files if indicated

    if args.retry_failed:
        retry_failed_files()

    # Start interactive shell if indicated

    if args.interactive:
        logger.info("Jumping to IPython shell.")
        import IPython
        _ns = dict(globals())
        _ns.update(locals())
        IPython.embed(header="Start IOP4ing!", module=sys.modules['__main__'], user_ns=_ns)

    sys.exit(0)



if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Script to reset the database, keeping the users and catalog data. USE WITH CARE."""

# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)

# other imports
import os
from datetime import datetime
from termcolor import colored, cprint

manage_fpath = f"{iop4conf.basedir}/iop4site/manage.py"
backupdir_path = f"{iop4conf.basedir}/priv.backups/"
datetime_str = datetime.now().strftime("%Y-%m-%d_%H%M")

print(colored("DANGER! This script will reset the database, keeping the users and catalog data. USE WITH CARE.", "yellow"))

if not os.path.exists(backupdir_path):
    os.makedirs(backupdir_path)

c = input("Are you sure you want to continue? (y/n) ")
if c != "y":
    print("Aborting.")
    exit(1)

print(f"Backing up catalog and users to {backupdir_path}/priv.iop4.dump.*.{datetime_str}.yaml ...")

os.system(f"python {manage_fpath} dumpdata --natural-primary --natural-foreign --format=yaml iop4api.astrosource > {backupdir_path}/priv.iop4.dump.catalog.{datetime_str}.yaml")
os.system(f"python {manage_fpath} dumpdata --natural-primary --natural-foreign --format=yaml auth > {backupdir_path}/priv.iop4.dump.auth.{datetime_str}.yaml")

c = input("Reset the DB? (y/n) ")
if c != "y":
    print("Aborting.")
    exit(1)

print("Resetting database ...")

os.system(rf"rm {iop4conf.db_path}")
os.system(rf"rm -rf {iop4conf.basedir}/iop4site/migrations")
os.system(rf"rm -rf {iop4conf.basedir}/iop4site/iop4api/migrations")
os.system(rf"python {manage_fpath} flush")
#os.system(r"python manage.py migrate iop4api zero")
os.system(rf"python {manage_fpath} makemigrations iop4api")
os.system(rf"python {manage_fpath} migrate")

print(f"Loading catalog and users from priv.iop4.dump.*.{datetime_str}.yaml ...")

c = input("Do you want to load the data? (y/n) ")
if c != "y":
    print("Aborting.")
    exit(1)

os.system(rf"python {manage_fpath} loaddata  {backupdir_path}/priv.iop4.*.{datetime_str}.yaml")


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from abc import ABCMeta, abstractmethod

# other imports
import re
import ftplib
import logging
import astropy.io.fits as fits
import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
import astrometry

# iop4lib imports
from iop4lib.enums import *
from .telescope import Telescope
from .osnt090 import OSNT090

# logging
import logging
logger = logging.getLogger(__name__)

class OSNT150(OSNT090, Telescope, metaclass=ABCMeta):
    """OSN-T150 telescope class.
    
    Inherits from OSNT090 since the remote archive access, reduction processes, etc are almost identical.
    """
    
    # telescope identification

    name = "OSN-T150"
    abbrv = "T150"
    telescop_kw = "T150-OSN"

    # telescope specific properties

    ftp_address = iop4conf.osn_t150_address
    ftp_user = iop4conf.osn_t150_user
    ftp_password = iop4conf.osn_t150_password
    ftp_encoding = 'utf-8'




# %% [markdown]
# # Details on astrometric calibration

# %%
# %run 01_notebook_configuration.py

# %% [markdown]
# The header of raw FITS files does not usually contain a valid WCS. The 
# astrometric calibration is implemented automatically during the 
# `ReducedFit.build_file()` step, and can be done separately through 
# `ReducedFit.astrometric_calibration()`.

# %%
from iop4lib.db import ReducedFit
redf = ReducedFit.by_fileloc("OSN-T090/2022-09-18/BLLac-0001R.fit")
redf.build_file()

# %% [markdown]
# For simple photometry images such as those from the Andor cameras, tries to
# detect sources in the image with different sets of parameters and 
# feeds them to a local astrometry.net solver. It tries with several sets until 
# it successfully calibrates the image. When done through `build_file()`, if 
# calibration is not achieved, it will give the ERROR_ASTROMETRY flag, 
# otherwise it will give it the BUILT_REDUCED flag.

# %%
redf

# %% 
redf.flag_labels

# %% [markdown]
# In the reduction log
# ```
# 2024-02-21 11:51 - pid 75152 - [plotting.py:476] - DEBUG - <ReducedFit 212 | OSN-T090/2022-09-18/BLLac-0001R.fit>: plotting astrometry summary image of background substraction results
# 2024-02-21 11:51 - pid 75152 - [plotting.py:488] - DEBUG - <ReducedFit 212 | OSN-T090/2022-09-18/BLLac-0001R.fit>: plotting astrometry summary image of segmentation results
# 2024-02-21 11:51 - pid 75152 - [plotting.py:582] - DEBUG - <ReducedFit 212 | OSN-T090/2022-09-18/BLLac-0001R.fit>: plotting astrometry summary image of astrometry results
# ``` 

# %% tags=["remove_input"]
from IPython.display import Markdown as md
md(f"you can see that several summary images were created. They are displayed in the [admin page for the reduced fit](/iop4/admin/iop4api/reducedfit/details/{redf.pk})")

# %% [markdown]
# For images with ordinary and extraordinary pairs such as those from CAFOS and 
# DIPOL, the reduction procedure first finds the sources and then divides them 
# in ordinaty and extraordinary lists. This is done through the utilities at
# `iop4lib.utils.sourcepairing` submodule.

# To find the pairs, the algorithm builds a distribution of the distance between
# all sources in the image, and selectes the most frequent distance with some
# resolution. This process is not free from error. It can be improved by 
# providing some bounds for this expected distance. For example, in a CAFOS 
# image such as

# %%
from iop4lib.db import RawFit
rf = RawFit.by_fileloc("CAHA-T220/2022-09-18/caf-20220918-23:01:33-sci-agui.fits")

# %% [markdown]
# We first create the reduced image and apply the calibration frames

# %%
redf = ReducedFit.create(rawfit=rf)
redf.apply_masters()

# %% [markdown]
# The `.has_pairs` property tells whether an image is expected to have pairs:

# %%
redf.has_pairs

# %% [markdown]
# Since we will be doing "blind" astrometric calibration with the local 
# astrometry.net solver, we will need (or at least greatly benefit) from hints
# for the size of the pixel and for pointing of the camera. These can be 
# obtained with 

# %%
size_hint = redf.get_astrometry_size_hint()
size_hint

# %% [markdown]
# This code to build this size hint is implemented in the corresponding 
# instrument class.

# %% [markdown]
# We will also benefit from a hint for the position of the object in the image.

# %% [markdown]
# The position is hint is built by the corresponding instrument class by looking
# at the RA and DEC (or equivalent) keywords in the FITS header.
# If the FITS header contains the RA and DEC (or equivalent) keywords, the 
# `header_hintcoord` property will return an SkyCoord object with the position

# %%
redf.header_hintcoord

# %% [markdown]

# This position should be close to the observed object. 
# The header_hintobject property returns the AstroSource in the catalog
# that matches the name in the OBJECT (or equivalent) keyword:

# %%
redf.header_hintobject

# %%
redf.header_hintobject.coord.separation(redf.header_hintcoord)

# %% [markdown]
# It can be be seen that the separation is indeed small.

# When header_hintcoord is available, 
# get_astrometry_position_hint() will return a hint for the object built from 
# it. Otherwise, it will look at the header_hintobject. If none are available, 
# it will raise an exception.

# %%
position_hint = redf.get_astrometry_position_hint()
position_hint

# %% [markdown]

# All these hints will be used to accelerate the blind calibration.

# Now, let's detect the stars visible in the image. The whole process will 
# depend on the parameters we select, and we might need to try with several 
# combinations. 

# %%
bkg_filter_size = 11 # a higher number will produce a more smooth bkg
bkg_box_size = 16 # a higher number will produce a more detailed bkg
# these bkg parameters should be set so the computed bkg does not include light from the point sources, but is detailed enough to show real variations along the image.
seg_kernel_size = None # none will automatically build a kernel of size 2n+1, which should be almost always a good approximation
npixels = 32 # related to the size of the sources in the image
seg_fwhm = 1.0 # the image will be convolved with a gaussian kernel of this width before src detection
n_rms_seg = 1.0 # the threshold of signal over background to count as a detection
# these detection parameters should return most of the real points in the image without too much fake sources
keep_n_seg = 200 # the maximum number of detected sources to use in the calibration
border_margin_px = 20 # remove any pixels less than this distance from the borders

# %% [markdown]
# We can use the utility functions in 
# `iop4lib.utils.sourcedetection` to get an bkg-estimation, which utilize the 
# `photutils` package (you could do the same, or try different methods following
# [its docs](https://photutils.readthedocs.io/en/stable/index.html).

# ### Background substraction

# %%
from iop4lib.utils.sourcedetection import get_bkg, get_segmentation, get_cat_sources_from_segment_map
bkg = get_bkg(redf.mdata, filter_size=bkg_filter_size, box_size=bkg_box_size)
imgdata_bkg_substracted = redf.mdata - bkg.background

from iop4lib.utils.plotting import plot_preview_background_substraction_1row
import matplotlib.pyplot as plt
import matplotlib as mplt
fig = plt.figure(figsize=(12, 5))
plot_preview_background_substraction_1row(redf, bkg, fig=fig)
plt.show()

# %% [markdown]
# ### Image Segmentation

# %%
seg_threshold = n_rms_seg * bkg.background_rms
segment_map, convolved_data = get_segmentation(imgdata_bkg_substracted, fwhm=seg_fwhm, kernel_size=seg_kernel_size, npixels=npixels, threshold=seg_threshold)
seg_cat, pos_seg, tb = get_cat_sources_from_segment_map(segment_map, imgdata_bkg_substracted, convolved_data)

# %%
from astropy.visualization import SqrtStretch, LogStretch, AsymmetricPercentileInterval
from astropy.visualization.mpl_normalize import ImageNormalize
fig, ax = plt.subplots()
ax.imshow(imgdata_bkg_substracted, cmap='gray', norm=ImageNormalize(imgdata_bkg_substracted.compressed(), stretch=LogStretch(a=10), interval=AsymmetricPercentileInterval(30, 99)))
seg_cat.plot_kron_apertures(color='r', lw=1.5)
plt.show()

# %% [markdown]
# ### Source pairing
# Now we need to distinguish between ordinary and extraordinary sources. To facilitate 
# this, you can use the `get_pairs_dxy` function, which will try to separate 
# them by looking at the most common distance between sources.

# First we find the most common distance between sources (both in x and y)

# %% 
from iop4lib.utils.sourcepairing import get_pairs_dxy, get_best_pairs
bins = 60
hist_range = (0, 60)
dx_min = dy_min = 0
dx_max = dy_max = 60
dx_eps = dy_eps = 1.0
seg1xy, seg2xy, seg_disp_xy, seg_disp_sign_xy = get_pairs_dxy(pos_seg, bins=bins, hist_range=hist_range, dx_min=dx_min, dx_max=dx_max, dy_min=dy_min, dy_max=dy_max, dx_eps=dx_eps, dy_eps=dy_eps, doplot=True)
seg_disp_sign_xy

# %% [markdown]
# Then we select the best pairs according to this distance, and recalculate it

# %%
seg1xy_best, seg2xy_best, seg_disp_xy_best, seg_disp_sign_xy_best = get_best_pairs(seg1xy, seg2xy, seg_disp_sign_xy)
seg_disp_sign_xy_best

# %% [markdown]
# We can plot the image again and check that indeed the pairs are well separated

# %%
from iop4lib.utils.plotting import imshow_w_sources
imshow_w_sources(redf.mdata, pos1=seg1xy_best, pos2=seg2xy_best)

# %% [markdown]
# To avoid mistakes, we provided lower and upper bounds for the distance between
# pairs. This can be specially important if there are few sources in the image.
# We could also do the same by ourselves, by looking at the distribution of
# distances:

# %%
import itertools
pairs = list(itertools.combinations(pos_seg, 2))
len(pairs)

# %%
dx_min = 0
dx_max = 60
dy_min = 0
dy_max = 60
paired = [(p1,p2) for p1,p2 in pairs if ( dx_min < abs( p1[0] - p2[0] ) < dx_max and dy_min < abs( p1[1] - p2[1] ) < dy_max )]
len(paired)

# %%
fig, ax = plt.subplots()
for i in [0,1]:
    distances = [abs(p1[i]-p2[i]) for p1,p2 in paired]
    plt.hist(distances, bins=60, range=(0,60), alpha=0.7)
plt.xlabel('Distance (px)')
plt.ylabel('Number of pairs')
plt.show()

# %% [markdown]
# Finally, we need to invoke the local astrometry.net solver to calibrate the image.

# %%
from iop4lib.utils.astrometry import solve_astrometry

solution = solve_astrometry(seg1xy_best, size_hint=size_hint, position_hint=position_hint)
solution

# %% [markdown]
# We can see that the solver was successful, and that the result has a logodds much
# higher than the default minimum of 21, meaning that the calibration was easy and trustable.

# The WCS (for one set of pairs) can be obtained with

# %%
from astropy.wcs import WCS
wcs1 = WCS(solution.best_match().wcs_fields)

# %% [markdown]
# We can plot the image with the sources and the WCS overlayed, and the catalog
# sources overplotted

# %%

fig, ax = plt.subplots(subplot_kw={'projection': wcs1})

imshow_w_sources(redf.mdata, ax=ax)

from iop4lib.db import AstroSource
from photutils.aperture import CircularAperture

for src in AstroSource.get_sources_in_field(wcs1, width=redf.mdata.shape[1], height=redf.mdata.shape[0]):
    ax.plot(*src.coord.to_pixel(wcs1), 'rx')
    ax.annotate(src.name, src.coord.to_pixel(wcs1), xytext=(20,0), textcoords='offset points', color='red', weight='bold')

ax.coords.grid(color='white', linestyle='solid')

ax.coords['ra'].set_axislabel('Right Ascension')
ax.coords['ra'].set_ticklabel_visible(True)
ax.coords['ra'].set_ticklabel_position('lb')

ax.coords['dec'].set_axislabel('Declination')
ax.coords['dec'].set_ticklabel_visible(True)
ax.coords['dec'].set_ticklabel_position('lb')

plt.show()

# %% [markdown]
# We can see that the sources are well aligned with the catalog sources. Since we
# already know the separation between the pairs, we can directly build the other 
# wcs by displacing the first one. However this will be done automatically when 
# we call the higher level methods in the `ReducedFit` class. In fact, the 
# `ReducedFit.astrometric_calibration()` method will do all the above and save 
# both WCS in the FITS header of the ReducedFit object:

# %%
redf.astrometric_calibration()

# %% 
redf.wcs1

# %%
redf.wcs2

# %% 
fig, ax = plt.subplots(subplot_kw={'projection': wcs1})

imshow_w_sources(redf.mdata, ax=ax)

from iop4lib.db import AstroSource
from photutils.aperture import CircularAperture

for src in AstroSource.get_sources_in_field(wcs1, width=redf.mdata.shape[1], height=redf.mdata.shape[0]):
    CircularAperture([*src.coord.to_pixel(redf.wcs1)], r=20).plot(ax=ax, color='r')
    ax.annotate(src.name, src.coord.to_pixel(redf.wcs1), xytext=(20,0), textcoords='offset points', color='red', weight='bold', verticalalignment='center')
    ax.plot(*src.coord.to_pixel(redf.wcs2), 'rx')

ax.coords.grid(color='white', linestyle='solid')

ax.coords['ra'].set_axislabel('Right Ascension')
ax.coords['ra'].set_ticklabel_visible(True)
ax.coords['ra'].set_ticklabel_position('bl')

ax.coords['dec'].set_axislabel('Declination')
ax.coords['dec'].set_ticklabel_visible(True)
ax.coords['dec'].set_ticklabel_position('lb')
#ax.coords['dec'].set_ticklabel(rotation='vertical')

plt.show()

# %% [markdown]
# If the calibration was successful, summary images containing more or less the 
# same information as the ones we plotted will be created and should be displayed 
# in the admin page for the reduced fit:

# %% tags=["remove_input"]
from IPython.display import Markdown as md
md(f"[/iop4/admin/iop4api/reducedfit/details/{redf.pk}](/iop4/admin/iop4api/reducedfit/details/{redf.pk})")


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.contrib import admin
from django.utils.html import format_html
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render, get_object_or_404
from django.http import QueryDict

# other imports
from iop4api.filters import *
from iop4api.models import *
from iop4admin import views

# iop4lib imports
from iop4lib.enums import *

# logging
import logging
logger = logging.getLogger(__name__)

class AdminAstroSource(admin.ModelAdmin):
    model = AstroSource
    list_display = ['name', 'other_name', 'ra_hms', 'dec_dms', 'srctype', 'get_last_reducedfit', 'get_calibrates', 'get_comment_firstline', 'get_details']
    search_fields = ['name', 'other_name', 'ra_hms', 'dec_dms', 'srctype', 'comment']
    list_filter = ('srctype',)
    
    @admin.display(description='CALIBRATES')
    def get_calibrates(self, obj):
        stars_str_L = obj.calibrates.all().values_list('name', flat=True)
        return "\n".join(stars_str_L)
    
    @admin.display(description='COMMENTS')
    def get_comment_firstline(self, obj):
        lines = obj.comment.split("\n")
        txt = lines[0]
        if len(lines) > 0 or len(lines[0]) > 30:
            txt = txt[:30] + "..."
        return txt
    
    @admin.display(description='LAST REDUCEDFIT')
    def get_last_reducedfit(self, obj):
        return obj.in_reducedfits.order_by('epoch__night').values_list('epoch__night', flat=True).last()
    
    @admin.display(description='DETAILS')
    def get_details(self, obj):
        url = reverse('iop4admin:iop4api_astrosource_details', args=[obj.id])
        return format_html(rf'<a href="{url}">Details</a>')
    
    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            path('details/<int:pk>', self.admin_site.admin_view(views.AstroSourceDetailsView.as_view()), name='iop4api_astrosource_details'),
        ]
        return my_urls + urls
    




# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# other imports

import itertools

# iop4lib

from iop4lib.db import Epoch
from .singleobj import SingleObjView

# logging

import logging
logger = logging.getLogger(__name__)

class EpochDetailsView(SingleObjView):
    model = Epoch
    template_name = "iop4admin/view_epochdetails.html"

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)

        obj = self.get_object()

        try:
            header_key_S = set(itertools.chain.from_iterable([rawfit.header.keys() for rawfit in obj.rawfits.all()]))
            context["header_key_S"] = list(header_key_S)
        except Exception as e:
            pass

        context["rawfitsummarystatus"] = obj.get_summary_rawfits_status()

        return context


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

 # django imports
from django.db import models

# other imports

# iop4lib imports
from ..enums import *

# logging
import logging
logger = logging.getLogger(__name__)



class AperPhotResult(models.Model):
    """ Aperture photometry results for an AstroSource in a ReducedFit, either the Ordinary or Extraordinary pair."""

    # database fields

    ## identifiers

    reducedfit = models.ForeignKey("ReducedFit", on_delete=models.CASCADE, related_name='aperphotresults', help_text="The ReducedFit this AperPhotResult has been computed for.")
    astrosource = models.ForeignKey("AstroSource", on_delete=models.CASCADE, related_name='aperphotresults', help_text="The AstroSource this AperPhotResult has been computed for.")
    aperpix = models.FloatField(null=True, blank=True)
    r_in = models.FloatField(null=True, blank=True)
    r_out = models.FloatField(null=True, blank=True)
    pairs = models.TextField(null=True, blank=True, choices=PAIRS.choices, help_text="Whether this AperPhotResult is for the Ordinary or Extraordinary pair.")

    ## photometry results

    bkg_flux_counts = models.FloatField(null=True, blank=True)
    bkg_flux_counts_err = models.FloatField(null=True, blank=True)
    
    flux_counts = models.FloatField(null=True, blank=True)
    flux_counts_err = models.FloatField(null=True, blank=True)

    ## extra fields
    fwhm = models.FloatField(null=True, blank=True)
    photopolresults = models.ManyToManyField("PhotoPolResult", related_name='aperphotresults', help_text="The PhotoPolResult(s) this AperPhotResult has been used for.")
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        app_label = 'iop4api'
        verbose_name = 'Aperture Photometry Result'
        verbose_name_plural = 'Aperture Photometry Results'
        constraints = [
            models.UniqueConstraint(fields=['reducedfit', 'astrosource', 'aperpix', 'r_in', 'r_out', 'pairs'], name='unique_aperphotresult')
        ]

  # repr and str

    def __repr__(self):
        return f'{self.__class__.__name__}.objects.get(id={self.id!r})'
    
    def __str__(self):
        return f'<{self.__class__.__name__} {self.id} | {self.reducedfit.fileloc} {self.astrosource.name} {self.aperpix} px {self.pairs}>' 

    @classmethod
    def create(cls, reducedfit, astrosource, aperpix, pairs, **kwargs):

        if (result := AperPhotResult.objects.filter(reducedfit=reducedfit, astrosource=astrosource, aperpix=aperpix, pairs=pairs).first()) is not None:
            logger.debug(f"AperPhotResult for {reducedfit}, {astrosource}, {aperpix}, {pairs} already exists, it will be used instead.")
        else:
            logger.debug(f"Creating AperPhotResult for {reducedfit}, {astrosource}, {aperpix}, {pairs}.")
            result = cls(reducedfit=reducedfit, astrosource=astrosource, aperpix=aperpix, pairs=pairs)

        for key, value in kwargs.items():
            setattr(result, key, value)

        result.save()




import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

from django.db import models

import os
import datetime
import numpy as np
import astropy.io.fits as fits

from .fitfilemodel import FitFileModel
from ..enums import *
from .fields import FlagChoices, FlagBitField

import logging
logger = logging.getLogger(__name__)

class MasterFlat(FitFileModel):
    """
    A class representing a master flat for an epoch.
    """

    margs_kwL = ['epoch', 'instrument', 'imgsize', 'band', 'obsmode', 'rotangle', 'exptime']

    imgtype = IMGTYPES.FLAT
    
    # Database fields
    rawfits = models.ManyToManyField('RawFit', related_name='built_masterflats')

    # flags

    class FLAGS(FlagChoices):
        IGNORE = 1 << 1
        
    flags = FlagBitField(choices=FLAGS.choices(), default=0)

    # fields corresponding to MasterFlat kw arguments (margs_kwL)

    epoch = models.ForeignKey('Epoch', on_delete=models.CASCADE, related_name='masterflats') 
    instrument = models.CharField(max_length=20, choices=INSTRUMENTS.choices)
    imgsize = models.CharField(max_length=12, null=True)
    band = models.CharField(max_length=10, choices=BANDS.choices, default=BANDS.NONE)
    obsmode = models.CharField(max_length=20, choices=OBSMODES.choices, default=OBSMODES.NONE)
    rotangle = models.FloatField(null=True)
    exptime = models.FloatField(null=True)

    masterbias = models.ForeignKey('MasterBias', null=True, on_delete=models.CASCADE, related_name='masterflats')
    masterdark = models.ForeignKey('MasterDark', null=True, on_delete=models.CASCADE, related_name='masterflats')

    class Meta:
        app_label = 'iop4api'
        verbose_name = "Master Flat"
        verbose_name_plural = "Master Flats"

    # Properties

    @property
    def filename(self):
        return f"masterflat_id{self.id}.fits"
    
    @property
    def filepath(self):
        return os.path.join(self.epoch.masterflatdir, self.filename)
    
    @property
    def fileloc(self):
        return f"{self.epoch.epochname}/{self.filename}"

    # Filed Properties location
    
    @property
    def filedpropdir(self):
       return os.path.join(self.epoch.calibrationdir, 
                            self.filename +  '.d', 
                            self.__class__.__name__)
    
    # some helper properties

    @property
    def margs(self):
        """
        Return a dict of the arguments used to build this MasterFlat.
        """
        return {'epoch':self.epoch, 'instrument':self.instrument, 'imgsize':self.imgsize, 'band':self.band, 'obsmode':self.obsmode, 'rotangle':self.rotangle, 'exptime':self.exptime}
    
    # repr and str

    @classmethod
    def margs2str(cls, margs):
       """
       Class method to build a nice string rep of the arguments of a MasterFlat.
       """
       return f"{margs['epoch'].epochname} | {margs['instrument']} | {margs['imgsize']} | {margs['band']} | {margs['obsmode']} | {margs['rotangle']} º | {margs['exptime']}s"

    def __repr__(self):
        return f"MasterFlat.objects.get(id={self.id!r})"
    
    def __str__(self):
        return f"<MasterFlat {self.id!r} | {MasterFlat.margs2str(self.margs)}>"
    
    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{self!r}")
        else:
            with p.group(4, f'<{self.__class__.__name__}(', ')>'):
                p.text(f"id: {self.id},")
                for k,v in self.margs.items():
                    p.breakable()
                    p.text(f"{k}: {v}")

    
    # Constructor
    @classmethod
    def create(cls, 
               rawfits=None, 
               masterbias=None,
               masterdark=None,
               auto_merge_to_db=True,
               force_rebuild=True,
               **kwargs):
        """
            Create a MasterFlat object for Epoch epoch.

            Parameters
            ----------
            epoch: Epoch
            instrument: str
            imgsize: "WidthxHeight" str
            band: str
            obsmode: str
            rotangle: float
            exptime: flat

            Other Parameters
            ----------------
            rawfits: list of RawFit (optional)
            masterbias: MasterBias (optional)
            auto_merge_to_db: bool (optional) Default: True
        """

        #margs = {k:kwargs[k] for k in MasterFlat.margs_kwL if k in kwargs}
        margs = {k:kwargs[k] for k in MasterFlat.margs_kwL} # so it gives error if some margs kw missing

        from iop4lib.db import RawFit

        if (mf := MasterFlat.objects.filter(**margs).first()) is not None:
            logger.info(f"{mf} exists, using it instead.")
        else:
            logger.info(f"A MasterFlat entry for {MasterFlat.margs2str(margs)} will be created.")
            mf = cls(**margs)
            logger.debug("Saving MasterFlat to DB so that it has an id.")
            mf.save()

        if rawfits is None:
            rawfits = RawFit.objects.filter(imgtype=IMGTYPES.FLAT, **margs)
            logger.debug(f"Found {len(rawfits)} flat raw files for {mf}.")
        else:
            logger.debug(f"Using {len(rawfits)} flat raw files for {mf}.")

        mf.rawfits.set(rawfits)

        if masterbias is None:
            from .masterbias import MasterBias
            margs = {k:kwargs[k] for k in MasterBias.margs_kwL if k in kwargs}
            masterbias = MasterBias.objects.filter(**margs).first()    
            logger.debug(f"Using {masterbias} as MasterBias for {mf}.")
            mf.masterbias = masterbias        

        if masterdark is None:
            from .masterdark import MasterDark
            margs = {k:kwargs[k] for k in MasterDark.margs_kwL if k in kwargs if k != 'exptime'} # exptime is a build parameter, but darks with different exptime can be used
            masterdark = MasterDark.objects.filter(**margs).first()    
            logger.debug(f"Using {masterdark} as MasterDark for {mf}.")
            mf.masterdark = masterdark

        # Build the file
        if not mf.fileexists or force_rebuild:
            logger.info(f"Building file")
            try:
                mf.build_file()
            except Exception as e:
                logger.error(f"An error ocurred while building the MasterFlat, deleting it and raising Exception.")
                mf.delete()
                raise Exception(f"An error ocurred while building the MasterFlat for {margs}: {e}")
            else:
                logger.info("MasterFlat created successfully.")

        # merge to DB

        if auto_merge_to_db:
            mf.save()

        return mf

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto_merge_to_db=True
    
    # Methods

    def build_file(self):
        logger.debug(f"Getting data from files")

        if self.rawfits.count() == 0:
            raise Exception(f"No rawfits for {self}")

        data_L = []
        for rf in self.rawfits.all():
            with fits.open(rf.filepath) as hdul:

                if self.masterdark is not None:
                    data = hdul[0].data - self.masterbias.data - self.masterdark.data * self.exptime
                else:
                    logger.warning(f"No MasterDark for {self}, is this a CCD and it is cold?")
                    data = (hdul[0].data - self.masterbias.data) 
                
                data = data / np.nanmedian(data)
                data_L.append(data)

        data = np.nanmedian(data_L, axis=0)
        data = data / np.nanmedian(data)

        logger.debug(f"Building header")

        header = fits.Header()

        header['TELESCOP'] = self.epoch.telescope
        header['NIGHT'] = self.epoch.yyyymmdd
        header['EPOCH'] = self.epoch.epochname
        header['IMGSIZE'] = self.imgsize
        header['IMGTYPE'] = 'masterflat'
        header['DATECREA'] = datetime.datetime.utcnow().isoformat(timespec="milliseconds")
        header['NRAWFITS'] = self.rawfits.count()
        header['BAND'] = self.band
        header['INSTRUME'] = self.instrument
        header['OBSMODE'] = self.obsmode
        header['ROTANGLE'] = self.rotangle
        header['EXPTIME'] = self.exptime

        logger.debug(f"Building HDU")

        if not os.path.isdir(self.epoch.masterflatdir):
            logger.debug(f"Creating directory {self.epoch.masterflatdir}")
            os.makedirs(self.epoch.masterflatdir)

        hdu = fits.PrimaryHDU(data, header=header)

        logger.debug(f"Writing MasterFlat to {self.filepath}")
        
        hdu.writeto(self.filepath, overwrite=True)


from django.contrib import admin
from iop4api.filters import *
from iop4api.models import *

# IOP4Admin custom admin

class IOP4AdminSite(admin.AdminSite):
    site_title = "IOP4 admin"
    site_header = 'IOP4 admin site'
    site_url = "/iop4/"
    index_title = 'Welcome to IOP4 admin site'
    
iop4admin_site = IOP4AdminSite(name='iop4admin')

# Register IOP4 models in the admin.

from .modeladmins.epoch import AdminEpoch
from .modeladmins.rawfit import AdminRawFit
from .modeladmins.masterflat import AdminMasterFlat
from .modeladmins.masterdark import AdminMasterDark
from .modeladmins.masterbias import AdminMasterBias
from .modeladmins.reducedfit import AdminReducedFit
from .modeladmins.astrosource import AdminAstroSource
from .modeladmins.aperphotresult import AdminAperPhotResult
from .modeladmins.photopolresult import AdminPhotoPolResult

iop4admin_site.register(Epoch, AdminEpoch)
iop4admin_site.register(RawFit, AdminRawFit)
iop4admin_site.register(MasterBias, AdminMasterBias)
iop4admin_site.register(MasterFlat, AdminMasterFlat)
iop4admin_site.register(MasterDark, AdminMasterDark)
iop4admin_site.register(ReducedFit, AdminReducedFit)
iop4admin_site.register(AstroSource, AdminAstroSource)
iop4admin_site.register(PhotoPolResult, AdminPhotoPolResult)
iop4admin_site.register(AperPhotResult, AdminAperPhotResult)




import iop4lib
iop4conf = iop4lib.Config(config_db=False)
from django.conf import settings
from django.http import Http404, FileResponse, HttpResponseForbidden, HttpResponseNotFound
from pathlib import Path
import mimetypes
import urllib

import logging
logger = logging.getLogger(__name__)

def docs(request, file_path=None):

    if file_path is None:
        file_path = "index.html"
    
    file_path = urllib.parse.unquote(file_path) 

    logger.debug(f"RECV: {file_path}")

    file_path = Path(iop4conf.basedir) / "docs" / "_build" / "html" / file_path

    if file_path.is_dir():
        file_path = file_path / "index.html"

    if settings.DEBUG:
        if file_path.exists():
            content_type = mimetypes.guess_type(file_path)[0]
            logger.debug(f"docs: {file_path} {content_type}")
            response = FileResponse(open(file_path, 'rb'), content_type=content_type)
            return response
        else:
            logger.debug(f"docs: {file_path} not found")
            return HttpResponseNotFound("Not found")
    else:
        return HttpResponseForbidden("Forbidden")


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.contrib import admin

from django.utils.html import format_html
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseNotFound, FileResponse
from django.apps import apps
from django.utils.safestring import mark_safe

# iop4lib imports
from iop4api.filters import *
from iop4api.models import *
from iop4admin.views import FitJS9View, FitDetailsView, get_fit_view
from iop4lib.enums import *

# other imports

# logging
import logging
logger = logging.getLogger(__name__)
    

@admin.action(description="Ignore these files in subsequent runs")
def action_mark_ignore(modeladmin, request, queryset):
    for obj in queryset:
        logger.debug(f"Setting IGNORE flag on {obj}")
        obj.set_flag(modeladmin.model.FLAGS.IGNORE)
        obj.save()

@admin.action(description="Do NOT ignore these files in subsequent runs")
def action_unmark_ignore(modeladmin, request, queryset):
    for obj in queryset:
        logger.debug(f"Removing IGNORE flag on {obj}")
        obj.unset_flag(modeladmin.model.FLAGS.IGNORE)
        obj.save()

class AdminFitFile(admin.ModelAdmin):

    list_per_page = 25
    list_max_show_all = 200
    
    @admin.display(description='TELESCOPE', ordering='epoch__telescope')
    def telescope(self, obj):
        return obj.epoch.telescope
    
    @admin.display(description='NIGHT', ordering='epoch__night')
    def night(self, obj):
        return obj.epoch.night
    
    @admin.display(description='FILENAME', ordering='filename')
    def filename(self, obj):
        return obj.filename

    @admin.display(description='STATUS')
    def status(self, obj):
        return ", ".join(obj.flag_labels)
    
    @admin.display(description="Built from")
    def get_built_from(self, obj):
        self.allow_tags = True
        
        ids_str_L = [str(rf.id) for rf in obj.rawfits.all()]
        a_href = reverse('iop4admin:%s_%s_changelist' % (RawFit._meta.app_label, RawFit._meta.model_name)) + "?id__in=%s" % ",".join(ids_str_L)
        a_text = ", ".join(ids_str_L)
        return mark_safe(f'<a href="{a_href}">{a_text}</a>')
    
    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            path('details/<int:pk>', self.admin_site.admin_view(get_fit_view(FitDetailsView,self.model).as_view()),  name=f"iop4api_{self.model._meta.model_name}_details"),
            path('viewer/<int:pk>', self.admin_site.admin_view(get_fit_view(FitJS9View,self.model).as_view()), name=f"iop4api_{self.model._meta.model_name}_viewer"),
            path('getfile/<int:pk>', self.admin_site.admin_view(self.view_getfile),  name=f"iop4api_{self.model._meta.model_name}_getfile"),
            path('preview/<int:pk>', self.admin_site.admin_view(self.view_preview),  name=f"iop4api_{self.model._meta.model_name}_preview"),
        ]
        return my_urls + urls
    
    def get_absolute_url(self):
        return reverse(f"iop4admin:{self.model._meta.app_label}_{self.model._meta.model_name}_details", args=[self.id])
    
    def view_getfile(self, request, pk):
        if ((fit := self.model.objects.filter(id=pk).first()) is None):
            return HttpResponseNotFound()

        if not fit.fileexists:
            return HttpResponseNotFound()

        return FileResponse(open(fit.filepath, 'rb'), filename=fit.filename)
    
    def view_preview(self, request, pk):
        
        if ((fit := self.model.objects.filter(id=pk).first()) is None):
            return HttpResponseNotFound()

        if not fit.fileexists:
            return HttpResponseNotFound()

        imgbytes = fit.get_imgbytes_preview_image()
        return HttpResponse(imgbytes, content_type="image/png")

    change_list_template = 'iop4admin/fits/change_list.html'
    
    def changelist_view(self, request, extra_context=None):
        # this gets the image preview checkbox value and leaves the url unaltered, otherwise
        # the url would redirect to ?e=1 because of the unexpected parameters
        request.GET._mutable = True
        toggle_image_preview = request.GET.pop('toggle_image_preview', False)
        request.GET._mutable = False

        # toggle the stored session option show_image_preview according to the GET parameter.

        if 'show_image_preview' not in request.session:
            show_image_preview = False
        else:
            show_image_preview = bool(request.session['show_image_preview'])

        if bool(toggle_image_preview):
            show_image_preview = not show_image_preview

        request.session['show_image_preview'] = show_image_preview

        # pass the current filters and page number to template so they get added to the url, otherwise they would be lost when toggling the cbox
        keep_in_url_dict = dict()
        filter_parameter_names = [f"{f}__exact" if isinstance(f,str) else f.parameter_name for f in self.get_list_filter(request)]
        keep_in_url_dict = {k: v for k, v in request.GET.items() if k in filter_parameter_names}
        if 'id' in request.GET: keep_in_url_dict['id'] = request.GET.get('id') # also id
        keep_in_url_dict['p'] = request.GET.get('p', 1) # also page number

        # build extra_context
        extra_context = extra_context or {}
        extra_context['keep_in_url_dict'] = keep_in_url_dict
        extra_context['show_image_preview'] = show_image_preview # will be used to display the cbox as checked or not

        return super().changelist_view(request, extra_context=extra_context)
    
    def get_list_display(self, request):
        list_display = list(super().get_list_display(request))

        if request.session.get('show_image_preview'):
            list_display = self.list_display + ['image_preview']

        return list_display
    
    def image_preview(self, obj, allow_tags=True):
        url_img_preview = reverse(f"iop4admin:iop4api_{self.model._meta.model_name}_preview", args=[obj.id])
        return format_html(rf"<img src='{url_img_preview}' width='64' height='64' />")


from django.db import models

from django.db.models import Lookup

from enum import Enum

class FlagChoices(Enum):

    def __add__(self, other):
        if isinstance(other, self.__class__):
            otherval = other.value
        else:
            otherval = other

        return self.value | otherval

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            otherval = other.value
        else:
            otherval = other

        if self.value & otherval != 0:
            return self.value & ~otherval
        else:
            return otherval
        
    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        return self.__sub__(other)

    @classmethod
    def choices(cls):
        return [(flag.value, flag.name) for flag in cls]

    @classmethod
    def get_label(cls, flag_value):
        """ Returns the label for the given flag value, does NOT allow for combinations of flags. 
        """
        if flag_value == 0:
            return '-'
        return cls(flag_value).name

    @classmethod
    def get_labels(cls, flag_value):
        """ Returns a list of labels for the given flag value, allows for combinations of flags. 
        """
        return [
            label
            for value, label in cls.choices()
            if value & flag_value != 0]


class FlagLookup(Lookup):
    """
    Any Lookup for this field should subclass this so it can accept the Enum directly.
    """
    def process_rhs(self, compiler, connection):
        if isinstance(self.rhs, Enum):
            self.rhs = self.rhs.value
        return super().process_rhs(compiler, connection)
    
class HasFlagLookup(FlagLookup):
    lookup_name = "has"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s & %s" % (lhs, rhs), params
    
class HasNotFlagLookup(FlagLookup):
    lookup_name = "hasnot"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "NOT (%s & %s)" % (lhs, rhs), params
    
class FlagPropertyDescriptor:
    def __init__(self, field):
        self.field = field

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__[self.field.attname]

    def __set__(self, instance, value):
        """ Allows to set it directly with a FlagChoice.FLAGNAME """
        if isinstance(value, Enum):
            value = value.value
        instance.__dict__[self.field.attname] = value

class FlagBitField(models.BigIntegerField):
    def get_prep_value(self, value):
        if isinstance(value, Enum):
            return value.value
        return super().get_prep_value(value)
    
    def __init__(self, choices=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.choices = choices
        self.register_lookup(HasFlagLookup)
        self.register_lookup(HasNotFlagLookup)

    def contribute_to_class(self, cls, name, **kwargs):
        """ 
        Provides methods to set, unset and check flags
        and give the object properties to retrieve flags as enum, 
        list of values and list of labels.
        """

        super().contribute_to_class(cls, name, **kwargs)
        setattr(cls, self.name, FlagPropertyDescriptor(self))

        def clear_flags(instance):
            instance.flags = 0
            return None
        
        def set_flag(instance, flag):
            if isinstance(flag, Enum):
                value = flag.value
            else:
                value = flag
            instance.flags |= value
            return None

        def unset_flag(instance, flag):
            if isinstance(flag, Enum):
                value = flag.value
            else:
                value = flag
            instance.flags &= ~value
            return None

        def has_flag(instance, flag):
            if isinstance(flag, Enum):
                value = flag.value
            else:
                value = flag
            return instance.flags & value != 0
        
        def has_not_flag(instance, flag):
            if isinstance(flag, Enum):
                value = flag.value
            else:
                value = flag
            return not (instance.flags & value != 0)
    
        @property
        def flag_list(instance):
            """ List of flags set, as enum choices (value, label). 
            """
            return [
                choice
                for choice in self.choices
                if instance.has_flag(choice[0])
            ]
        
        @property
        def flag_values(instance):
            """ list of values for the flags set. 
            """
            return [
                value
                for value, label in self.choices
                if instance.has_flag(value)
            ]
        
        @property
        def flag_labels(instance):
            """ List of labels for the flags set. 
            """
            return [
                label
                for value, label in self.choices
                if instance.has_flag(value)
            ]
        
        # methods for flag managing

        cls.clear_flags = clear_flags
        cls.set_flag = set_flag
        cls.unset_flag = unset_flag
        cls.has_flag = has_flag
        cls.has_not_flag = has_not_flag

        # properties for flag retrieving

        cls.flag_list = flag_list
        cls.flag_values = flag_values
        cls.flag_labels = flag_labels

        # method for multiple flag checking
        
        def has_flags(instance, flags):
            """ Return True if all flags given are set. 
            """
            return all([instance.has_flag(flag) for flag in flags])
        
        def has_not_flags(instance, flags):
            """ Return True if all flags given are NOT set.
            """
            return all([instance.has_not_flag(flag) for flag in flags])
    
        cls.has_flags = has_flags
        cls.has_not_flags = has_not_flags
        



import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

from django.db import models

from abc import ABCMeta, abstractmethod

import os
import io
import base64
import numpy as np
import scipy as sp
import scipy.stats as stats
import matplotlib as mplt
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from astropy.visualization.mpl_normalize import ImageNormalize
from astropy.visualization import SqrtStretch, LogStretch, AsymmetricPercentileInterval
import astropy.io.fits as fits
from iop4lib.utils.filedproperty import FiledProperty, filed_property

import logging
logger = logging.getLogger(__name__)

class AbstractModelMeta(ABCMeta, type(models.Model)):
    pass

class AbstractModel(models.Model, metaclass=AbstractModelMeta):    
    class Meta:
        abstract = True


class FitFileModel(AbstractModel):
    """
    TODO: this should be an abstract class (abc) and model which can be inherited.
    This class adds some common functionality for fit files (get image preview, get histrogram, statistics, etc)
    """

    class Meta:
        abstract = True

    # Abstract properties

    # Identity

    @property
    @abstractmethod
    def epoch(self):
        raise NotImplementedError
    
    @property
    @abstractmethod
    def filename(self):
        raise NotImplementedError
    
    @property
    @abstractmethod
    def fileloc(self):
        raise NotImplementedError

    # File paths

    @property
    @abstractmethod
    def filepath(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def filedpropdir(self):
        raise NotImplementedError

    # Helper properties

    @property
    def fileexists(self):
        return os.path.isfile(self.filepath)

    @property
    def header(self):
        """
        Returns header.
        """        
        return fits.getheader(self.filepath)

    @property
    def data(self):
        """
        Returns data.
        """
        return fits.getdata(self.filepath)

    @property
    def mdata(self):
        """
        Return masked data (finite values).
        """
        data = fits.getdata(self.filepath)
        return np.ma.array(data, mask=~np.isfinite(data))

    @property
    def width(self):
        """
        Returns the width of the image.
        """
        return self.header['NAXIS1']
    
    @property
    def height(self):
        """
        Returns the height of the image.
        """
        return self.header['NAXIS2']

    # Common helper properties

    @property
    def stats(self):
        """
        Returns a dict with some important statistics about the data (mean, etc).
        """
        
        if not os.path.exists(self.filepath):
            return dict()
        
        from iop4lib.utils import stats_dict

        with fits.open(self.filepath) as hdul:
            data = hdul[0].data.flatten()

        data = data[np.isfinite(data)]

        return stats_dict(data)
        
    # Common helper methods for visualizing data in the web admin

    def get_imgbytes_preview_image(self, force_rebuild=False, **kwargs):
        """
        Build an image preview (png) of the data.

        Parameters
        ----------
        normtype : str
            'log', 'logstretch', or a matplotlib norm object.
        a : float
            If norm is 'logstretch', this is the stretch factor.
        vmin, vmax: float
            If norm is 'log', these are the min and max values for the norm.


        If called with default arguments (no kwargs) it will try to load from the disk,
        except if called with force_rebuild.

        When called with default arguments (no kwargs), if rebuilt, it will save the image to disk.
        """

        # logger.debug(f"Requested image preview for {self.fileloc}, {kwargs=}")
                     
        imgdata = self.mdata

        width = kwargs.get('width', 256)
        height = kwargs.get('height', 256)
        normtype = kwargs.get('norm', "log")
        vmin = kwargs.get('vmin', np.quantile(imgdata.compressed(), 0.3))
        vmax = kwargs.get('vmax', np.quantile(imgdata.compressed(), 0.99))
        a = kwargs.get('a', 10)

        fpath = os.path.join(self.filedpropdir, "img_preview_image.png")

        if len(kwargs) == 0 and not force_rebuild:
            if os.path.isfile(fpath) and os.path.getmtime(self.filepath) < os.path.getmtime(fpath):
                # logger.debug(f"Loading image preview for {self.fileloc} from disk")
                with open(fpath, 'rb') as f:
                    return f.read()

        # logger.debug(f"Building image preview for {self.fileloc}")

        cmap = plt.cm.gray.copy()
        cmap.set_bad(color='red')
        cmap.set_under(color='black')
        cmap.set_over(color='white')
            
        if normtype == "log":
            norm = ImageNormalize(imgdata, vmin=vmin, vmax=vmax, stretch=LogStretch(a=a))
        elif normtype == "logstretch":
            norm = ImageNormalize(stretch=LogStretch(a=a))


        buf = io.BytesIO()

        fig = mplt.figure.Figure(figsize=(width/100, height/100), dpi=iop4conf.mplt_default_dpi)
        ax = fig.subplots()
        ax.imshow(imgdata, cmap=cmap, origin='lower', norm=norm)
        try:
            from iop4lib.db import ReducedFit, AstroSource
            from iop4lib.enums import SRCTYPES
            # If it is a astro calibrated reduced fit, mark the src position
            if self.has_flag(ReducedFit.FLAGS.BUILT_REDUCED):
                if (target_src := self.sources_in_field.exclude(srctype=SRCTYPES.CALIBRATOR).get()) is not None:
                    target_pos_px = target_src.coord.to_pixel(self.wcs1)
                    ax.axhline(y=target_pos_px[1], color='r', linestyle="--", linewidth=1)
                    ax.axvline(x=target_pos_px[0], color='r', linestyle="--", linewidth=1)
        except Exception as e: 
            # it can fail if there is any problem with the fit calibration (e.g. in early versions the wcs
            # was not saved into key A, but we still want to be able to explore the data in the admin).
            logger.warning(f"Coudl not mark the target source position on ReducedFit {self.pk}: {e}")

        ax.axis('off')
        fig.savefig(buf, format='png', bbox_inches='tight', pad_inches=0)
        fig.clf()

        # logger.debug(f"Image preview for {self.fileloc} built")

        buf.seek(0)
        imgbytes = buf.read()

        # if it was rebuilt, save it to disk if it is the default image settings.
        if len(kwargs) == 0:
            # logger.debug(f"Saving image preview for {self.fileloc} to disk")
            if not os.path.exists(self.filedpropdir):
                os.makedirs(self.filedpropdir)
            with open(fpath, 'wb') as f:
                f.write(imgbytes)

        return imgbytes
        
            
    def get_imgbytes_preview_histogram(self, force_rebuild=False, **kwargs):
        """
        If called with default arguments (no kwargs) it will try to load from the disk,
        except if called with force_rebuild.

        When called with default arguments (no kwargs), if rebuilt, it will save the image to disk.
        """

        # logger.debug(f"Requested preview histogram for {self.fileloc}, {kwargs=}")

        width = kwargs.get('width', 256)
        height = kwargs.get('height', 120)
        xscale = kwargs.get('xscale', "log")
        yscale = kwargs.get('yscale', "log")
        a = kwargs.get('a', 10)

        fpath = os.path.join(self.filedpropdir, "img_preview_histogram.png")

        if len(kwargs) == 0 and not force_rebuild:
            if os.path.isfile(fpath) and os.path.getmtime(self.filepath) < os.path.getmtime(fpath):
                #logger.debug(f"Loading preview histogram for {self.fileloc} from disk")
                with open(fpath, 'rb') as f:
                    return f.read()
         
        # logger.debug(f"Building preview histogram for {self.fileloc}")

        imgdata = self.mdata.flatten()

        buf = io.BytesIO()

        fig = mplt.figure.Figure(figsize=(width/100, height/100), dpi=iop4conf.mplt_default_dpi)
        ax = fig.subplots()

        ax.hist(imgdata, bins='sqrt', log=True, histtype='step')

        for q in np.quantile(imgdata, [0.5]):
            ax.axvline(x=q, color='r', linestyle="--", linewidth=1, label=f"{q:.2f}")

        # set axes    
        if xscale == "log":        
            vabsmin = np.amin(imgdata)
            vabsmax = np.amax(imgdata)
                
            a = 10

            def fscale(x):
                xs = (x - vabsmin) / (vabsmax - vabsmin)
                y = np.log(a * xs + 1) / np.log(a + 1)
                return y

            def fscale_inv(y):
                xs = (np.exp(y * np.log(a+1)) - 1) / a
                x = xs * (vabsmax - vabsmin) + vabsmin
                return x
            
            ax.set_xscale('function', functions=(fscale, fscale_inv))
            ax.set_xlim([vabsmin, vabsmax])  
        else:
            raise Exception("Not expecting xscale other than log")

        ax.set_yscale(yscale)

        # turn off axis
        ax.axis('off')

        fig.savefig(buf, format='png', bbox_inches='tight', pad_inches=0)
        fig.clf()

        # logger.debug(f"Preview histogram for {self.fileloc} built")

        buf.seek(0)
        imgbytes = buf.read()
        
        # if it was rebuilt, save it to disk if it is the default image settings.
        if len(kwargs) == 0:
            # logger.debug(f"Saving preview histogram for {self.fileloc} to disk")
            if not os.path.exists(self.filedpropdir):
                os.makedirs(self.filedpropdir)
            with open(fpath, 'wb') as f:
                f.write(imgbytes)
                    
        return imgbytes
        






""" # how it would be done without matplotlib
imgdata = LogNorm(vmin=vmin, vmax=vmax)(imgdata)

imgdata = 256*imgdata
imgdata = imgdata.astype(np.uint8)
image = Image.fromarray(imgdata).resize((width,height))


buf = io.BytesIO()
image.save(buf, format='png')
buf.seek(0)
imgbytes = buf.read()
"""


from django.contrib import admin

from django.utils.html import format_html
from django.urls import reverse 
from django.utils.safestring import mark_safe

from iop4api.filters import *
from iop4api.models import *
from .fitfile import AdminFitFile, action_mark_ignore, action_unmark_ignore

import logging
logger = logging.getLogger(__name__)

class AdminMasterDark(AdminFitFile):
    
    model = MasterDark

    list_display = ['id', 'telescope', 'night', 'instrument', 'imgsize', 'exptime', 'get_masterbias', 'get_built_from', 'options', 'status']
    
    list_filter = (
            RawFitIdFilter,
            RawFitTelescopeFilter,
            RawFitNightFilter,
            RawFitInstrumentFilter,
            RawFitFlagFilter,
            "imgsize",
    )

    actions = [action_mark_ignore, action_unmark_ignore]

    @admin.display(description='Options')
    def options(self, obj):
        url_details = reverse('iop4admin:iop4api_masterdark_details', args=[obj.id])
        url_viewer= reverse('iop4admin:iop4api_masterdark_details', args=[obj.id])
        return format_html(rf'<a href="{url_details}">details</a> / <a href="{url_viewer}">advanced viewer</a>')
    
    @admin.display(description='Telescope')
    def telescope(self, obj):
        return obj.epoch.telescope
    
    @admin.display(description='Night')
    def night(self, obj):
        return obj.epoch.night
    
    @admin.display(description='MasterBias')
    def get_masterbias(self, obj):
        self.allow_tags = True
        if obj.masterbias is None:
            return "-"
        url = reverse('iop4admin:%s_%s_changelist' % (MasterBias._meta.app_label, MasterBias._meta.model_name)) + f"?id={obj.masterbias.id}"
        return mark_safe(rf'<a href="{url}">{obj.masterbias.id}</a>')
    


from django.contrib import admin

from django.utils.html import format_html
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render

from iop4api.filters import *
from iop4api.models import *
from .fitfile import AdminFitFile

import logging
logger = logging.getLogger(__name__)

class AdminRawFit(AdminFitFile):
    model = RawFit
    list_display = ["id", 'filename', 'telescope', 'night', 'instrument', 'status', 'imgtype', 'imgsize', 'band', 'obsmode', 'rotangle', 'exptime', 'options']
    readonly_fields = [field.name for field in RawFit._meta.fields] 
    search_fields = ['id', 'filename', 'epoch__telescope', 'epoch__night'] 
    ordering = ['-epoch__night','-epoch__telescope']
    list_filter = (
            RawFitIdFilter,
            RawFitTelescopeFilter,
            RawFitNightFilter,
            RawFitFilenameFilter,
            RawFitInstrumentFilter,
            RawFitFlagFilter,
            "imgtype",
            "obsmode",
            "band",
            "imgsize",
        )
    
    
    @admin.display(description='OPTIONS')
    def options(self, obj):
        html_src = str()

        if obj.imgtype == IMGTYPES.LIGHT and hasattr(obj, "reduced"):
            url_reduced = reverse('iop4admin:%s_%s_changelist' % (ReducedFit._meta.app_label, ReducedFit._meta.model_name)) + f"?id={obj.reduced.id}"
            html_src += rf'<a href="{url_reduced}">reduced</a> / '
        
        url_details = reverse('iop4admin:iop4api_rawfit_details', args=[obj.id])
        url_viewer= reverse('iop4admin:iop4api_rawfit_viewer', args=[obj.id])
        html_src += rf'<a href="{url_details}">details</a> / <a href="{url_viewer}">advanced viewer</a>'

        return format_html(html_src)

    def image_preview(self, obj, allow_tags=True):
        url_img_preview = reverse('iop4admin:iop4api_rawfit_preview', args=[obj.id])
        return format_html(rf"<img src='{url_img_preview}' width='64' height='64' />")